/*
 * Copyright (c) 2006 Ondrej Palkovsky
 * Copyright (c) 2008 Martin Decky
 * Copyright (c) 2011 Petr Koupy
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup kgraph
 * @{
 */
/**
 * @file
 */

#include <sys/types.h>
#include <errno.h>
#include <adt/list.h>
#include <io/mode.h>
#include <io/charfieldmap.h>
#include <libgraph.h>
#include "../kgraph.h"
#include "serial.h"

#define SERIAL_WIDTH  80
#define SERIAL_HEIGHT 24

vslmode_list_element_t mode;

static int serial_claim(visualizer_t *vs)
{
	vt100_state_t *state = (vt100_state_t *) vs->dev_ctx;
	return vt100_claim(state);
}

static int serial_yield(visualizer_t *vs)
{
	vt100_state_t *state = (vt100_state_t *) vs->dev_ctx;
	return vt100_yield(state);
}

static int serial_change_mode(visualizer_t *vs, vslmode_t new_mode)
{
	return EOK;
}

static int serial_handle_damage(visualizer_t *vs,
    sysarg_t col0, sysarg_t row0, sysarg_t cols, sysarg_t rows,
    sysarg_t col_offset, sysarg_t row_offset)
{
	vt100_state_t *state = (vt100_state_t *) vs->dev_ctx;
	charfieldmap_t *map = &vs->cells.f;
	charfield_t *field;

	for (sysarg_t row = row0; row < rows + row0; ++row) {
		for (sysarg_t col = col0; col < cols + col0; ++col) {
			field = charfieldmap_charfield_at(map,
			    (col + col_offset) % map->cols,
			    (row + row_offset) % map->rows);
			vt100_goto(state, col, row);
			vt100_set_attr(state, field->attrs);
			vt100_putchar(state, field->ch);
		}
	}

	return EOK;
}

static int serial_suspend(visualizer_t *vs)
{
	return EOK;
}

static int serial_wakeup(visualizer_t *vs)
{
	return EOK;
}

static visualizer_ops_t serial_ops = {
	.claim = serial_claim,
	.yield = serial_yield,
	.change_mode = serial_change_mode,
	.handle_damage = serial_handle_damage,
	.suspend = serial_suspend,
	.wakeup = serial_wakeup
};

int serial_init(vt100_putchar_t putchar_fn, vt100_control_puts_t control_puts_fn, ddf_dev_t *dev)
{
	int rc;

	vt100_state_t *state =
	    vt100_state_create(SERIAL_WIDTH, SERIAL_HEIGHT, putchar_fn, control_puts_fn);
	if (state == NULL)
		return ENOMEM;

	ddf_fun_t *fun_vs = ddf_fun_create(dev, fun_exposed, "vsl0");
	if (fun_vs == NULL) {
		vt100_state_destroy(state);
		return ENOMEM;
	}
	fun_vs->ops = &graph_vsl_device_ops;

	visualizer_t *vs = ddf_fun_data_alloc(fun_vs, sizeof(visualizer_t));
	if (vs == NULL) {
		vt100_state_destroy(state);
		return ENOMEM;
	}
	graph_init_visualizer(vs);

	mode.mode.index = 0;
	mode.mode.version = 0;
	mode.mode.refresh_rate = 0;
	mode.mode.screen_aspect.width = SERIAL_WIDTH;
	mode.mode.screen_aspect.height = SERIAL_HEIGHT;
	mode.mode.screen_width = SERIAL_WIDTH;
	mode.mode.screen_height = SERIAL_HEIGHT;
	mode.mode.cell_type = CELL_TYPE_FIELD;
	mode.mode.cell_aspect.width = 1;
	mode.mode.cell_aspect.height = 1;
	mode.mode.cell_visual.field_visual = CHAR_ATTR_INDEX;

	link_initialize(&mode.link);
	list_append(&mode.link, &vs->modes);

	vs->def_mode_idx = 0;

	vs->ops = serial_ops;
	vs->dev_ctx = (void *)state;

	rc = ddf_fun_bind(fun_vs);
	if (rc != EOK) {
		list_remove(&mode.link);
		vt100_state_destroy(state);
		vs->dev_ctx = NULL;
		ddf_fun_destroy(fun_vs);
		return rc;
	}

	vs->reg_svc_handle = fun_vs->handle;
	ddf_fun_add_to_category(fun_vs, "visualizer");

	return EOK;
}

/** @}
 */
