/*
 * Copyright (c) 2011 Martin Decky
 * Copyright (c) 2011 Petr Koupy
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup kgraph
 * @{
 */
/**
 * @file
 */

#include <sys/types.h>
#include <errno.h>

#include <malloc.h>
#include <as.h>
#include <align.h>

#include <sysinfo.h>
#include <ddi.h>
#include <libarch/ddi.h>
#include <task.h>

#include <adt/list.h>
#include <str.h>

#include <io/mode.h>
#include <io/charfieldmap.h>

#include <libgraph.h>

#include "../kgraph.h"
#include "ega.h"

#define EGA_IO_BASE  ((ioport8_t *) 0x3d4)
#define EGA_IO_SIZE  2

#define FB_POS(x, y)  (((y) * ega.width + (x)) << 1)

typedef struct {
	sysarg_t width;
	sysarg_t height;
	
	uint8_t style_normal;
	uint8_t style_inverted;

	size_t size;
	uint8_t *addr;
} ega_t;

static ega_t ega;

static vslmode_list_element_t mode;

static inline uint8_t attrs_attr(char_attrs_t attrs)
{
	uint8_t attr = 0;
	
	switch (attrs.type) {
	case CHAR_ATTR_STYLE:
		switch (attrs.val.style) {
		case STYLE_NORMAL:
			attr = ega.style_normal;
			break;
		case STYLE_EMPHASIS:
			attr = (ega.style_normal | 0x04);
			break;
		case STYLE_INVERTED:
			attr = ega.style_inverted;
			break;
		case STYLE_SELECTED:
			attr = (ega.style_inverted | 0x40);
			break;
		}
		break;
	case CHAR_ATTR_INDEX:
		attr = ((attrs.val.index.bgcolor & 7) << 4) |
		    (attrs.val.index.fgcolor & 7);
		if (attrs.val.index.attr & CATTR_BRIGHT)
			attr |= 0x08;
		break;
	case CHAR_ATTR_RGB:
		attr = (attrs.val.rgb.bgcolor < attrs.val.rgb.fgcolor) ?
		    ega.style_inverted : ega.style_normal;
		break;
	}
	
	return attr;
}

static int ega_claim(visualizer_t *vs)
{
	return EOK;
}

static int ega_yield(visualizer_t *vs)
{
	return EOK;
}

static int ega_change_mode(visualizer_t *vs, vslmode_t new_mode)
{
	return EOK;
}

static int ega_handle_damage(visualizer_t *vs,
    sysarg_t col0, sysarg_t row0, sysarg_t cols, sysarg_t rows,
    sysarg_t col_offset, sysarg_t row_offset)
{
	charfieldmap_t *map = &vs->cells.f;
	charfield_t *field;
	uint8_t glyph;

	for (sysarg_t row = row0; row < rows + row0; ++row) {
		for (sysarg_t col = col0; col < cols + col0; ++col) {
			field = charfieldmap_charfield_at(map, 
			    (col + col_offset) % map->cols,
			    (row + row_offset) % map->rows);

			if (ascii_check(field->ch))
				glyph = field->ch;
			else
				glyph = '?';

			uint8_t attr = attrs_attr(field->attrs);

			ega.addr[FB_POS(col, row)] = glyph;
			ega.addr[FB_POS(col, row) + 1] = attr;
		}
	}

	return EOK;
}

static int ega_suspend(visualizer_t *vs)
{
	return EOK;
}

static int ega_wakeup(visualizer_t *vs)
{
	return EOK;
}

static visualizer_ops_t ega_ops = {
	.claim = ega_claim,
	.yield = ega_yield,
	.change_mode = ega_change_mode,
	.handle_damage = ega_handle_damage,
	.suspend = ega_suspend,
	.wakeup = ega_wakeup
};

int ega_init(ddf_dev_t *dev)
{
	sysarg_t present;
	int rc = sysinfo_get_value("fb", &present);
	if (rc != EOK)
		present = false;
	
	if (!present)
		return ENOENT;
	
	sysarg_t kind;
	rc = sysinfo_get_value("fb.kind", &kind);
	if (rc != EOK)
		kind = (sysarg_t) -1;
	
	if (kind != 2)
		return EINVAL;
	
	sysarg_t paddr;
	rc = sysinfo_get_value("fb.address.physical", &paddr);
	if (rc != EOK)
		return rc;
	
	sysarg_t width;
	rc = sysinfo_get_value("fb.width", &width);
	if (rc != EOK)
		return rc;
	
	sysarg_t height;
	rc = sysinfo_get_value("fb.height", &height);
	if (rc != EOK)
		return rc;
	
	rc = iospace_enable(task_get_id(), (void *) EGA_IO_BASE, EGA_IO_SIZE);
	if (rc != EOK)
		return rc;
	
	ega.width = width;
	ega.height = height;
	
	ega.size = (width * height) << 1;
	rc = physmem_map((void *) paddr,
	    ALIGN_UP(ega.size, PAGE_SIZE) >> PAGE_WIDTH,
	    AS_AREA_READ | AS_AREA_WRITE, (void *) &ega.addr);
	if (rc != EOK)
		return rc;
	
	sysarg_t blinking;
	rc = sysinfo_get_value("fb.blinking", &blinking);
	if (rc != EOK)
		blinking = false;
	
	ega.style_normal = 0xf0;
	ega.style_inverted = 0x0f;
	
	if (blinking) {
		ega.style_normal &= 0x77;
		ega.style_inverted &= 0x77;
	}
	
	ddf_fun_t *fun_vs = ddf_fun_create(dev, fun_exposed, "vsl0");
	if (fun_vs == NULL) {
		as_area_destroy(ega.addr);
		return ENOMEM;
	}
	fun_vs->ops = &graph_vsl_device_ops;

	visualizer_t *vs = ddf_fun_data_alloc(fun_vs, sizeof(visualizer_t));
	if (vs == NULL) {
		as_area_destroy(ega.addr);
		return ENOMEM;
	}
	graph_init_visualizer(vs);

	mode.mode.index = 0;
	mode.mode.version = 0;
	mode.mode.refresh_rate = 0;
	mode.mode.screen_aspect.width = width;
	mode.mode.screen_aspect.height = height;
	mode.mode.screen_width = width;
	mode.mode.screen_height = height;
	mode.mode.cell_type = CELL_TYPE_FIELD;
	mode.mode.cell_aspect.width = 1;
	mode.mode.cell_aspect.height = 1;
	mode.mode.cell_visual.field_visual = CHAR_ATTR_INDEX;

	link_initialize(&mode.link);
	list_append(&mode.link, &vs->modes);

	vs->def_mode_idx = 0;

	vs->ops = ega_ops;
	vs->dev_ctx = NULL;

	rc = ddf_fun_bind(fun_vs);
	if (rc != EOK) {
		list_remove(&mode.link);
		ddf_fun_destroy(fun_vs);
		as_area_destroy(ega.addr);
		return rc;
	}

	vs->reg_svc_handle = fun_vs->handle;
	ddf_fun_add_to_category(fun_vs, "visualizer");
	
	return EOK;
}

/** @}
 */
