/*
 * Copyright (c) 2006 Jakub Vana
 * Copyright (c) 2006 Ondrej Palkovsky
 * Copyright (c) 2008 Martin Decky
 * Copyright (c) 2011 Petr Koupy
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup kgraph
 * @{
 */
/**
 * @file
 */

#include <abi/fb/visuals.h>
#include <sys/types.h>
#include <errno.h>

#include <malloc.h>
#include <mem.h>
#include <as.h>
#include <align.h>

#include <sysinfo.h>
#include <ddi.h>

#include <adt/list.h>

#include <io/mode.h>
#include <io/pixelmap.h>
#include <io/charfieldmap.h>

#include <pixconv.h>

#include <libgraph.h>

#include "../gfx/font-8x16.h"
#include "../kgraph.h"
#include "kfb.h"

#define FB_POS(x, y)  ((y) * kfb.scanline + (x) * kfb.pixel_bytes)

#define COL2X(col)  ((col) * FONT_WIDTH)
#define ROW2Y(row)  ((row) * FONT_SCANLINES)
#define X2COL(x)    ((x) / FONT_WIDTH)
#define Y2ROW(y)    ((y) / FONT_SCANLINES)

#define GLYPH_POS(glyph, y) \
	((glyph) * (kfb.glyph_bytes) + (y) * (kfb.glyph_scanline))

typedef void (*draw_char_t)(sysarg_t, sysarg_t, wchar_t, pixel_t, pixel_t);

typedef struct {
	sysarg_t width;
	sysarg_t height;
	size_t offset;
	size_t scanline;
	visual_t visual;
	
	pixel2visual_t pixel2visual;
	visual2pixel_t visual2pixel;
	visual_mask_t visual_mask;
	size_t pixel_bytes;

	draw_char_t draw_char;
	size_t glyph_scanline;
	size_t glyph_bytes;
	uint8_t *glyphs;

	size_t size;
	uint8_t *addr;
} kfb_t;

static kfb_t kfb;

static vslmode_list_element_t field_mode;
static vslmode_list_element_t pixel_mode;

static pixel_t color_table[16] = {
	[COLOR_BLACK]       = 0x000000,
	[COLOR_BLUE]        = 0x0000f0,
	[COLOR_GREEN]       = 0x00f000,
	[COLOR_CYAN]        = 0x00f0f0,
	[COLOR_RED]         = 0xf00000,
	[COLOR_MAGENTA]     = 0xf000f0,
	[COLOR_YELLOW]      = 0xf0f000,
	[COLOR_WHITE]       = 0xf0f0f0,

	[COLOR_BLACK + 8]   = 0x000000,
	[COLOR_BLUE + 8]    = 0x0000ff,
	[COLOR_GREEN + 8]   = 0x00ff00,
	[COLOR_CYAN + 8]    = 0x00ffff,
	[COLOR_RED + 8]     = 0xff0000,
	[COLOR_MAGENTA + 8] = 0xff00ff,
	[COLOR_YELLOW + 8]  = 0xffff00,
	[COLOR_WHITE + 8]   = 0xffffff,
};

static inline void attrs_rgb(char_attrs_t attrs, pixel_t *bgcolor, pixel_t *fgcolor)
{
	switch (attrs.type) {
	case CHAR_ATTR_STYLE:
		switch (attrs.val.style) {
		case STYLE_NORMAL:
			*bgcolor = color_table[COLOR_WHITE];
			*fgcolor = color_table[COLOR_BLACK];
			break;
		case STYLE_EMPHASIS:
			*bgcolor = color_table[COLOR_WHITE];
			*fgcolor = color_table[COLOR_RED];
			break;
		case STYLE_INVERTED:
			*bgcolor = color_table[COLOR_BLACK];
			*fgcolor = color_table[COLOR_WHITE];
			break;
		case STYLE_SELECTED:
			*bgcolor = color_table[COLOR_RED];
			*fgcolor = color_table[COLOR_WHITE];
			break;
		}
		break;
	case CHAR_ATTR_INDEX:
		*bgcolor = color_table[(attrs.val.index.bgcolor & 7) |
		    ((attrs.val.index.attr & CATTR_BRIGHT) ? 8 : 0)];
		*fgcolor = color_table[(attrs.val.index.fgcolor & 7) |
		    ((attrs.val.index.attr & CATTR_BRIGHT) ? 8 : 0)];
		break;
	case CHAR_ATTR_RGB:
		*bgcolor = attrs.val.rgb.bgcolor;
		*fgcolor = attrs.val.rgb.fgcolor;
		break;
	}
}

/** Draw a character, takes advantage of alignment.
 *
 * This version can only be used if the following conditions are met:
 *
 *   - word size is divisible by pixel_bytes
 *   - cell scanline size is divisible by word size
 *   - cell scanlines are word-aligned
 *
 * It makes use of the pre-rendered mask to process (possibly) several
 * pixels at once (word size / pixel_bytes pixels at a time are processed)
 * making it very fast. Most notably this version is not applicable at 24 bits
 * per pixel.
 *
 * @param x        X coordinate of top-left corner on screen.
 * @param y        Y coordinate of top-left corner on screen.
 * @param ch       Character to draw.
 * @param bgcolor  Background color.
 * @param fbgcolor Foreground color.
 */
static void draw_char_aligned(sysarg_t x, sysarg_t y, wchar_t ch,
    pixel_t bgcolor, pixel_t fgcolor)
{
	size_t word_size = sizeof(unsigned long);
	
	/* Prepare a pair of words, one filled with foreground-color
	 * pattern and the other filled with background-color pattern. */
	unsigned long fg_buf;
	unsigned long bg_buf;
	
	for (size_t i = 0; i < word_size / kfb.pixel_bytes; i++) {
		kfb.pixel2visual(&((uint8_t *) &bg_buf)[i * kfb.pixel_bytes], bgcolor);
		kfb.pixel2visual(&((uint8_t *) &fg_buf)[i * kfb.pixel_bytes], fgcolor);
	}
	
	/* Pointer to the current position in the mask. */
	unsigned long *maskp =
	    (unsigned long *) &kfb.glyphs[GLYPH_POS(fb_font_glyph(ch), 0)];
	
	/* Pointer to the current position on the screen. */
	unsigned long *dst = (unsigned long *) &kfb.addr[FB_POS(x, y)];
	
	/* Width of the character cell in words. */
	size_t ww = FONT_WIDTH * kfb.pixel_bytes / word_size;
	
	/* Offset to add when moving to another screen scanline. */
	size_t d_add = kfb.scanline - FONT_WIDTH * kfb.pixel_bytes;
	
	for (size_t yd = 0; yd < FONT_SCANLINES; yd++) {
		/* Now process the cell scanline, combining foreground
		 * and background color patters using the pre-rendered mask. */
		for (size_t i = 0; i < ww; i++) {
			unsigned long mask = *maskp++;
			*dst++ = (fg_buf & mask) | (bg_buf & ~mask);
		}
		
		/* Move to the beginning of the next scanline of the cell. */
		dst = (unsigned long *) ((uint8_t *) dst + d_add);
	}
}

/** Draw a character, fallback version.
 *
 * This version does not make use of the pre-rendered mask, it uses
 * the font bitmap directly. It works always, but it is slower.
 *
 * @param x        X coordinate of top-left corner on screen.
 * @param y        Y coordinate of top-left corner on screen.
 * @param ch       Character to draw.
 * @param bgcolor  Background color.
 * @param fgcolor  Foreground color.
 */
static void draw_char_fallback(sysarg_t x, sysarg_t y, wchar_t ch,
    pixel_t bgcolor, pixel_t fgcolor)
{
	/* Character glyph */
	uint16_t glyph = fb_font_glyph(ch);
	
	/* Pre-render the foreground and background color pixels. */
	uint8_t fg_buf[4];
	uint8_t bg_buf[4];
	
	kfb.pixel2visual(bg_buf, bgcolor);
	kfb.pixel2visual(fg_buf, fgcolor);
	
	/* Pointer to the current position on the screen. */
	uint8_t *dst = (uint8_t *) &kfb.addr[FB_POS(x, y)];
	
	/* Offset to add when moving to another screen scanline. */
	size_t d_add = kfb.scanline - FONT_WIDTH * kfb.pixel_bytes;
	
	for (size_t yd = 0; yd < FONT_SCANLINES; yd++) {
		/* Byte containing bits of the glyph scanline. */
		uint8_t byte = fb_font[glyph][yd];
		
		for (size_t i = 0; i < FONT_WIDTH; i++) {
			/* Choose color based on the current bit. */
			uint8_t *src = (byte & 0x80) ? fg_buf : bg_buf;
			
			/* Copy the pixel. */
			for (size_t j = 0; j < kfb.pixel_bytes; j++)
				*dst++ = *src++;
			
			/* Move to the next bit. */
			byte <<= 1;
		}
		
		/* Move to the beginning of the next scanline of the cell. */
		dst += d_add;
	}
}

static int kfb_claim(visualizer_t *vs)
{
	return EOK;
}

static int kfb_yield(visualizer_t *vs)
{
	if (vs->mode_set) {
		vs->ops.handle_damage = NULL;
	}

	return EOK;
}

static int kfb_handle_damage_fields(visualizer_t *vs,
    sysarg_t col0, sysarg_t row0, sysarg_t cols, sysarg_t rows,
    sysarg_t col_offset, sysarg_t row_offset)
{
	charfieldmap_t *map = &vs->cells.f;
	charfield_t *field;
	pixel_t bgcolor = 0;
	pixel_t fgcolor = 0;

	/* If pixel dimensions are not divisible by field dimensions, fill the
	 * resulting narrow stripes on the right and bottom edge of the screen. */
	bool full_damage = (cols == map->cols) && (rows == map->rows);
	if (full_damage) {
		sysarg_t horiz_begin = 
		    (field_mode.mode.screen_width * field_mode.mode.cell_aspect.width);
		sysarg_t horiz_end = field_mode.mode.screen_aspect.width;
		sysarg_t vert_begin = 
		    (field_mode.mode.screen_height * field_mode.mode.cell_aspect.height);
		sysarg_t vert_end = field_mode.mode.screen_aspect.height;
		for (sysarg_t y = 0; y < field_mode.mode.screen_aspect.height; ++y) {
			for (sysarg_t x = horiz_begin; x < horiz_end; ++x) {
				kfb.pixel2visual(kfb.addr + FB_POS(x, y), color_table[COLOR_WHITE]);
			}
		}
		for (sysarg_t y = vert_begin; y < vert_end; ++y) {
			for (sysarg_t x = 0; x < field_mode.mode.screen_aspect.width; ++x) {
				kfb.pixel2visual(kfb.addr + FB_POS(x, y), color_table[COLOR_WHITE]);
			}
		}
	}

	for (sysarg_t row = row0; row < rows + row0; ++row) {
		for (sysarg_t col = col0; col < cols + col0; ++col) {
			field = charfieldmap_charfield_at(map, 
			    (col + col_offset) % map->cols,
			    (row + row_offset) % map->rows);
			attrs_rgb(field->attrs, &bgcolor, &fgcolor);
			kfb.draw_char(COL2X(col), ROW2Y(row), field->ch, bgcolor, fgcolor);
		}
	}

	return EOK;
}

static int kfb_handle_damage_pixels(visualizer_t *vs,
    sysarg_t x0, sysarg_t y0, sysarg_t width, sysarg_t height,
    sysarg_t x_offset, sysarg_t y_offset)
{
	pixelmap_t *map = &vs->cells.p;

	if (x_offset == 0 && y_offset == 0) {
		/* Faster damage routine ignoring offsets. */
		for (sysarg_t y = y0; y < height + y0; ++y) {
			for (sysarg_t x = x0; x < width + x0; ++x) {
				kfb.pixel2visual(kfb.addr + FB_POS(x, y),
				    *pixelmap_pixel_at(map, x, y));
			}
		}
	} else {
		for (sysarg_t y = y0; y < height + y0; ++y) {
			for (sysarg_t x = x0; x < width + x0; ++x) {
				kfb.pixel2visual(kfb.addr + FB_POS(x, y),
				    *pixelmap_pixel_at(map,
				    (x + x_offset) % map->width,
				    (y + y_offset) % map->height));
			}
		}
	}

	return EOK;
}

static int kfb_change_mode(visualizer_t *vs, vslmode_t new_mode)
{
	switch (new_mode.cell_type) {
	case CELL_TYPE_FIELD:
		vs->ops.handle_damage = kfb_handle_damage_fields;
		return EOK;
	case CELL_TYPE_PIXEL:
		vs->ops.handle_damage = kfb_handle_damage_pixels;
		return EOK;
	default:
		return EINVAL;
	}
}

static int kfb_suspend(visualizer_t *vs)
{
	return EOK;
}

static int kfb_wakeup(visualizer_t *vs)
{
	return EOK;
}

static visualizer_ops_t kfb_ops = {
	.claim = kfb_claim,
	.yield = kfb_yield,
	.change_mode = kfb_change_mode,
	.handle_damage = NULL,
	.suspend = kfb_suspend,
	.wakeup = kfb_wakeup
};

/** Render glyphs
 *
 * Convert glyphs from device independent font
 * description to current visual representation.
 */
static void render_glyphs(size_t sz)
{
	memset(kfb.glyphs, 0, sz);
	
	for (unsigned int glyph = 0; glyph < FONT_GLYPHS; glyph++) {
		for (unsigned int y = 0; y < FONT_SCANLINES; y++) {
			for (unsigned int x = 0; x < FONT_WIDTH; x++) {
				kfb.visual_mask(
				    &kfb.glyphs[GLYPH_POS(glyph, y) + x * kfb.pixel_bytes],
				    (fb_font[glyph][y] & (1 << (7 - x))) ? true : false);
			}
		}
	}
}

int kfb_init(ddf_dev_t *dev)
{
	sysarg_t present;
	int rc = sysinfo_get_value("fb", &present);
	if (rc != EOK)
		present = false;

	if (!present)
		return ENOENT;
	
	sysarg_t kind;
	rc = sysinfo_get_value("fb.kind", &kind);
	if (rc != EOK)
		kind = (sysarg_t) -1;
	
	if (kind != 1)
		return EINVAL;

	sysarg_t paddr;
	rc = sysinfo_get_value("fb.address.physical", &paddr);
	if (rc != EOK)
		return rc;
	
	sysarg_t offset;
	rc = sysinfo_get_value("fb.offset", &offset);
	if (rc != EOK)
		offset = 0;
	
	sysarg_t width;
	rc = sysinfo_get_value("fb.width", &width);
	if (rc != EOK)
		return rc;
	
	sysarg_t height;
	rc = sysinfo_get_value("fb.height", &height);
	if (rc != EOK)
		return rc;
	
	sysarg_t scanline;
	rc = sysinfo_get_value("fb.scanline", &scanline);
	if (rc != EOK)
		return rc;
	
	sysarg_t visual;
	rc = sysinfo_get_value("fb.visual", &visual);
	if (rc != EOK)
		return rc;
	
	kfb.width = width;
	kfb.height = height;
	kfb.offset = offset;
	kfb.scanline = scanline;
	kfb.visual = visual;

	switch (visual) {
	case VISUAL_INDIRECT_8:
		kfb.pixel2visual = pixel2bgr_323;
		kfb.visual2pixel = bgr_323_2pixel;
		kfb.visual_mask = visual_mask_323;
		kfb.pixel_bytes = 1;
		break;
	case VISUAL_RGB_5_5_5_LE:
		kfb.pixel2visual = pixel2rgb_555_le;
		kfb.visual2pixel = rgb_555_le_2pixel;
		kfb.visual_mask = visual_mask_555;
		kfb.pixel_bytes = 2;
		break;
	case VISUAL_RGB_5_5_5_BE:
		kfb.pixel2visual = pixel2rgb_555_be;
		kfb.visual2pixel = rgb_555_be_2pixel;
		kfb.visual_mask = visual_mask_555;
		kfb.pixel_bytes = 2;
		break;
	case VISUAL_RGB_5_6_5_LE:
		kfb.pixel2visual = pixel2rgb_565_le;
		kfb.visual2pixel = rgb_565_le_2pixel;
		kfb.visual_mask = visual_mask_565;
		kfb.pixel_bytes = 2;
		break;
	case VISUAL_RGB_5_6_5_BE:
		kfb.pixel2visual = pixel2rgb_565_be;
		kfb.visual2pixel = rgb_565_be_2pixel;
		kfb.visual_mask = visual_mask_565;
		kfb.pixel_bytes = 2;
		break;
	case VISUAL_RGB_8_8_8:
		kfb.pixel2visual = pixel2rgb_888;
		kfb.visual2pixel = rgb_888_2pixel;
		kfb.visual_mask = visual_mask_888;
		kfb.pixel_bytes = 3;
		break;
	case VISUAL_BGR_8_8_8:
		kfb.pixel2visual = pixel2bgr_888;
		kfb.visual2pixel = bgr_888_2pixel;
		kfb.visual_mask = visual_mask_888;
		kfb.pixel_bytes = 3;
		break;
	case VISUAL_RGB_8_8_8_0:
		kfb.pixel2visual = pixel2rgb_8880;
		kfb.visual2pixel = rgb_8880_2pixel;
		kfb.visual_mask = visual_mask_8880;
		kfb.pixel_bytes = 4;
		break;
	case VISUAL_RGB_0_8_8_8:
		kfb.pixel2visual = pixel2rgb_0888;
		kfb.visual2pixel = rgb_0888_2pixel;
		kfb.visual_mask = visual_mask_0888;
		kfb.pixel_bytes = 4;
		break;
	case VISUAL_BGR_0_8_8_8:
		kfb.pixel2visual = pixel2bgr_0888;
		kfb.visual2pixel = bgr_0888_2pixel;
		kfb.visual_mask = visual_mask_0888;
		kfb.pixel_bytes = 4;
		break;
	case VISUAL_BGR_8_8_8_0:
		kfb.pixel2visual = pixel2bgr_8880;
		kfb.visual2pixel = bgr_8880_2pixel;
		kfb.visual_mask = visual_mask_8880;
		kfb.pixel_bytes = 4;
		break;
	default:
		return EINVAL;
	}
	
	kfb.glyph_scanline = FONT_WIDTH * kfb.pixel_bytes;
	kfb.glyph_bytes = kfb.glyph_scanline * FONT_SCANLINES;
	
	size_t sz = FONT_GLYPHS * kfb.glyph_bytes;
	kfb.glyphs = (uint8_t *) malloc(sz);
	if (kfb.glyphs == NULL)
		return EINVAL;
	
	render_glyphs(sz);

	kfb.size = scanline * height;
	rc = physmem_map((void *) paddr + offset,
	    ALIGN_UP(kfb.size, PAGE_SIZE) >> PAGE_WIDTH,
	    AS_AREA_READ | AS_AREA_WRITE, (void *) &kfb.addr);
	if (rc != EOK) {
		free(kfb.glyphs);
		return rc;
	}

	/* Conditions necessary to select aligned glyph drawing variants:
	 *  - word size is divisible by number of bytes per pixel
	 *  - cell scanline size is divisible by word size
	 *  - cell scanlines are word-aligned */
	size_t word_size = sizeof(unsigned long);
	if (((word_size % kfb.pixel_bytes) == 0)
	    && ((FONT_WIDTH * kfb.pixel_bytes) % word_size == 0)
	    && (kfb.scanline % word_size == 0)) {
		kfb.draw_char = draw_char_aligned;
	} else {
		kfb.draw_char = draw_char_fallback;
	}

	ddf_fun_t *fun_vs = ddf_fun_create(dev, fun_exposed, "vsl0");
	if (fun_vs == NULL) {
		free(kfb.glyphs);
		as_area_destroy(kfb.addr);
		return ENOMEM;
	}
	fun_vs->ops = &graph_vsl_device_ops;

	visualizer_t *vs = ddf_fun_data_alloc(fun_vs, sizeof(visualizer_t));
	if (vs == NULL) {
		free(kfb.glyphs);
		as_area_destroy(kfb.addr);
		return ENOMEM;
	}
	graph_init_visualizer(vs);

	field_mode.mode.index = 0;
	field_mode.mode.version = 0;
	field_mode.mode.refresh_rate = 0;
	field_mode.mode.screen_aspect.width = width;
	field_mode.mode.screen_aspect.height = height;
	field_mode.mode.screen_width = width / FONT_WIDTH;
	field_mode.mode.screen_height = height / FONT_SCANLINES;
	field_mode.mode.cell_type = CELL_TYPE_FIELD;
	field_mode.mode.cell_aspect.width = FONT_WIDTH;
	field_mode.mode.cell_aspect.height = FONT_SCANLINES;
	field_mode.mode.cell_visual.field_visual = CHAR_ATTR_RGB;

	pixel_mode.mode.index = 1;
	pixel_mode.mode.version = 0;
	pixel_mode.mode.refresh_rate = 0;
	pixel_mode.mode.screen_aspect.width = width;
	pixel_mode.mode.screen_aspect.height = height;
	pixel_mode.mode.screen_width = width;
	pixel_mode.mode.screen_height = height;
	pixel_mode.mode.cell_type = CELL_TYPE_PIXEL;
	pixel_mode.mode.cell_aspect.width = 1;
	pixel_mode.mode.cell_aspect.height = 1;
	pixel_mode.mode.cell_visual.pixel_visual = visual;

	link_initialize(&field_mode.link);
	list_append(&field_mode.link, &vs->modes);
	link_initialize(&pixel_mode.link);
	list_append(&pixel_mode.link, &vs->modes);

#ifdef CONFIG_PREFER_PIXEL_MODE
	vs->def_mode_idx = 1;
#else
	vs->def_mode_idx = 0;
#endif

	vs->ops = kfb_ops;
	vs->dev_ctx = NULL;

	rc = ddf_fun_bind(fun_vs);
	if (rc != EOK) {
		list_remove(&field_mode.link);
		list_remove(&pixel_mode.link);
		ddf_fun_destroy(fun_vs);
		free(kfb.glyphs);
		as_area_destroy(kfb.addr);
		return rc;
	}

	vs->reg_svc_handle = fun_vs->handle;
	ddf_fun_add_to_category(fun_vs, "visualizer");

	return EOK;
}

/** @}
 */
