/*
 * Copyright (c) 2012 Petr Koupy
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup gui
 * @{
 */
/**
 * @file
 */

#include <errno.h>
#include <malloc.h>
#include <io/charfieldmap.h>
#include <ipc/input.h>

#include <surface.h>
#include <gfx/font-8x16.h>

#include "window.h"
#include "terminal.h"

#define NAME               "emuterm"
#define NAMESPACE          "inputemu"
#define RELAY_NAME         "rly"

#define LOCFS_MOUNT_POINT  "/loc"
#define SRV_CONSOLE        "/srv/console"
#define APP_GETTERM        "/app/getterm"
#define APP_BDSH           "/app/bdsh"
#define APP_KLOG           "/app/klog"

static sysarg_t namespace_idx = 0;
static sysarg_t relay_idx = 0;
static LIST_INITIALIZE(relay_list);
static FIBRIL_MUTEX_INITIALIZE(relay_list_mtx);

static pixel_t color_table[16] = {
	[COLOR_BLACK]       = PIXEL(255, 0, 0, 0),
	[COLOR_BLUE]        = PIXEL(255, 0, 0, 240),
	[COLOR_GREEN]       = PIXEL(255, 0, 240, 0),
	[COLOR_CYAN]        = PIXEL(255, 0, 240, 240),
	[COLOR_RED]         = PIXEL(255, 240, 0, 0),
	[COLOR_MAGENTA]     = PIXEL(255, 240, 0, 240),
	[COLOR_YELLOW]      = PIXEL(255, 240, 240, 0),
	[COLOR_WHITE]       = PIXEL(255, 240, 240, 240),

	[COLOR_BLACK + 8]   = PIXEL(255, 0, 0, 0),
	[COLOR_BLUE + 8]    = PIXEL(255, 0, 0, 255),
	[COLOR_GREEN + 8]   = PIXEL(255, 0, 255, 0),
	[COLOR_CYAN + 8]    = PIXEL(255, 0, 255, 255),
	[COLOR_RED + 8]     = PIXEL(255, 255, 0, 0),
	[COLOR_MAGENTA + 8] = PIXEL(255, 255, 0, 255),
	[COLOR_YELLOW + 8]  = PIXEL(255, 255, 255, 0),
	[COLOR_WHITE + 8]   = PIXEL(255, 255, 255, 255),
};

static inline void attrs_rgb(char_attrs_t attrs, pixel_t *bgcolor, pixel_t *fgcolor)
{
	switch (attrs.type) {
	case CHAR_ATTR_STYLE:
		switch (attrs.val.style) {
		case STYLE_NORMAL:
			*bgcolor = color_table[COLOR_WHITE];
			*fgcolor = color_table[COLOR_BLACK];
			break;
		case STYLE_EMPHASIS:
			*bgcolor = color_table[COLOR_WHITE];
			*fgcolor = color_table[COLOR_RED];
			break;
		case STYLE_INVERTED:
			*bgcolor = color_table[COLOR_BLACK];
			*fgcolor = color_table[COLOR_WHITE];
			break;
		case STYLE_SELECTED:
			*bgcolor = color_table[COLOR_RED];
			*fgcolor = color_table[COLOR_WHITE];
			break;
		}
		break;
	case CHAR_ATTR_INDEX:
		*bgcolor = color_table[(attrs.val.index.bgcolor & 7) |
		    ((attrs.val.index.attr & CATTR_BRIGHT) ? 8 : 0)];
		*fgcolor = color_table[(attrs.val.index.fgcolor & 7) |
		    ((attrs.val.index.attr & CATTR_BRIGHT) ? 8 : 0)];
		break;
	case CHAR_ATTR_RGB:
		*bgcolor = 0xff000000 | attrs.val.rgb.bgcolor;
		*fgcolor = 0xff000000 | attrs.val.rgb.fgcolor;
		break;
	}
}

static void draw_char_internal(sysarg_t hpos, sysarg_t vpos, wchar_t ch,
    pixel_t bgcolor, pixel_t fgcolor, surface_t *surface)
{
	uint16_t glyph = fb_font_glyph(ch);
	for (unsigned int y = 0; y < FONT_SCANLINES; ++y) {
		for (unsigned int x = 0; x < FONT_WIDTH; ++x) {
			pixel_t p = (fb_font[glyph][y] & (1 << (7 - x))) ?
			    fgcolor : bgcolor;
			surface_put_pixel(surface, hpos + x, vpos + y, p);
		}
	}
}

static int emu_claim(visualizer_t *vs)
{
	return EOK;
}

static int emu_yield(visualizer_t *vs)
{
	return EOK;
}

static int emu_handle_damage(visualizer_t *vs,
    sysarg_t col0, sysarg_t row0, sysarg_t cols, sysarg_t rows,
    sysarg_t col_offset, sysarg_t row_offset)
{
	terminal_t *term = (terminal_t *) vs->dev_ctx;

	/* Claim buffer or window surface. */
	surface_t *surface;
	sysarg_t hpos, vpos;
	if (term->buffered) {
		hpos = 0;
		vpos = 0;
		fibril_mutex_lock(&term->buffer_mtx);
		surface = term->buffer;
		if (!surface) {
			fibril_mutex_unlock(&term->buffer_mtx);
			return EOK;
		}
	} else {
		hpos = term->widget.hpos;
		vpos = term->widget.vpos;
		surface = window_claim(term->widget.window);
		if (!surface) {
			window_yield(term->widget.window);
			return EOK;
		}
	}

	charfieldmap_t *map = &vs->cells.f;
	pixel_t white = PIXEL(255, 240, 240, 240);
	pixel_t bgcolor = 0;
	pixel_t fgcolor = 0;

	/* If pixel dimensions are not divisible by field dimensions, fill the
	 * resulting narrow stripes on the right and bottom edge of the terminal. */
	bool full_damage = (cols == map->cols) && (rows == map->rows);
	if (full_damage) {
		vslmode_t *mode = &term->mode.mode;
		sysarg_t horiz_begin = hpos + (mode->screen_width * mode->cell_aspect.width);
		sysarg_t horiz_end = hpos + mode->screen_aspect.width;
		sysarg_t vert_begin = vpos + (mode->screen_height * mode->cell_aspect.height);
		sysarg_t vert_end = vpos + mode->screen_aspect.height;
		for (sysarg_t y = vpos; y < vert_end; ++y) {
			for (sysarg_t x = horiz_begin; x < horiz_end; ++x) {
				surface_put_pixel(surface, x, y, white);
			}
		}
		for (sysarg_t y = vert_begin; y < vert_end; ++y) {
			for (sysarg_t x = hpos; x < horiz_end; ++x) {
				surface_put_pixel(surface, x, y, white);
			}
		}
	}

	/* Handle console damage. */
	for (sysarg_t row = row0; row < rows + row0; ++row) {
		for (sysarg_t col = col0; col < cols + col0; ++col) {
			charfield_t *field = charfieldmap_charfield_at(map,
			    (col + col_offset) % map->cols,
			    (row + row_offset) % map->rows);
			sysarg_t x = hpos + (col * FONT_WIDTH);
			sysarg_t y = vpos + (row * FONT_SCANLINES);
			attrs_rgb(field->attrs, &bgcolor, &fgcolor);
			draw_char_internal(x, y, field->ch, bgcolor, fgcolor, surface);
		}
	}

	/* Yield claimed surface. */
	if (term->buffered) {
		fibril_mutex_unlock(&term->buffer_mtx);
		window_refresh(term->widget.window);
	} else {
		window_yield(term->widget.window);
		window_damage(term->widget.window);
	}

	return EOK;
}

static int emu_change_mode(visualizer_t *vs, vslmode_t new_mode)
{
	return EOK;
}

static int emu_suspend(visualizer_t *vs)
{
	return EOK;
}

static int emu_wakeup(visualizer_t *vs)
{
	return EOK;
}

static visualizer_ops_t emu_ops = {
	.claim = emu_claim,
	.yield = emu_yield,
	.change_mode = emu_change_mode,
	.handle_damage = emu_handle_damage,
	.suspend = emu_suspend,
	.wakeup = emu_wakeup
};

static int input_find_namespace_idx(void)
{
	int rc = EEXISTS;
	while (rc == EEXISTS) {
		char probe[LOC_NAME_MAXLEN + 1];
		snprintf(probe, LOC_NAME_MAXLEN, "%s/%zu", NAMESPACE, namespace_idx);
		service_id_t dummy_dsid;

		rc = loc_service_register(probe, &dummy_dsid);
		if (rc != EOK && rc != EEXISTS) {
			return rc;
		}

		if (rc == EEXISTS) {
			namespace_idx++;
		} else {
			/* Unused index found. */
		}
	}
	return rc;
}

static void input_init_relay(terminal_input_relay_t *rly)
{
	link_initialize(&rly->link);
	rly->sess = NULL;
}

static int input_register_relay(terminal_input_relay_t *rly)
{
	int rc = EOK;

	char node[LOC_NAME_MAXLEN + 1];
	snprintf(node, LOC_NAME_MAXLEN, "%s%zu/%s%zu", NAMESPACE,
	    namespace_idx, RELAY_NAME, relay_idx++);

	rc = loc_service_register(node, &rly->dsid);
	if (rc != EOK) {
		return rc;
	}

	fibril_mutex_lock(&relay_list_mtx);
	list_append(&rly->link, &relay_list);
	fibril_mutex_unlock(&relay_list_mtx);

	return rc;
}

static terminal_input_relay_t *input_get_relay(service_id_t dsid)
{
	terminal_input_relay_t *relay = NULL;

	fibril_mutex_lock(&relay_list_mtx);
	list_foreach(relay_list, link) {
		terminal_input_relay_t *cur =
		    list_get_instance(link, terminal_input_relay_t, link);
		if (cur->dsid == dsid) {
			relay = cur;
			break;
		}
	}
	fibril_mutex_unlock(&relay_list_mtx);

	return relay;
}

static int input_unregister_relay(terminal_input_relay_t *rly)
{
	int rc = EOK;

	fibril_mutex_lock(&relay_list_mtx);
	rc = loc_service_unregister(rly->dsid);
	list_remove(&rly->link);
	fibril_mutex_unlock(&relay_list_mtx);

	return rc;
}

void deinit_terminal(terminal_t *term)
{
	async_hangup(term->relay.sess);
	graph_notify_disconnect(
	    term->visualizer->notif_sess, term->visualizer->reg_svc_handle);
	task_kill(term->console_task);
	input_unregister_relay(&term->relay);
	graph_unregister_visualizer(term->visualizer);
	term->visualizer->dev_ctx = NULL;
	list_remove(&term->mode.link);
	graph_destroy_visualizer(term->visualizer);
	widget_deinit(&term->widget);
	if (term->buffered && term->buffer) {
		surface_destroy(term->buffer);
	}
}

static void terminal_destroy(widget_t *widget)
{
	terminal_t *term = (terminal_t *) widget;

	deinit_terminal(term);

	free(term);
}

static void terminal_reconfigure(widget_t *widget)
{
	/* no-op */
}

static bool copy_buffer_internal(terminal_t *term)
{
	/* Copy buffer onto window surface. */
	bool damage = false;
	if (term->buffer) {
		sysarg_t hpos = term->widget.hpos;
		sysarg_t vpos = term->widget.vpos;
		surface_t *surface = window_claim(term->widget.window);
		if (surface) {
			sysarg_t _x, _y, width, height;
			fibril_mutex_lock(&term->buffer_mtx);
			surface_get_damaged_region(term->buffer, &_x, &_y, &width, &height);
			damage = (width > 0) && (height > 0);
			for (sysarg_t y = _y; y < _y + height; ++y) {
				for (sysarg_t x = _x; x < _x + width; ++x) {
					pixel_t pix = surface_get_pixel(term->buffer, x, y);
					surface_put_pixel(surface, hpos + x, vpos + y, pix);
				}
			}
			surface_reset_damaged_region(term->buffer);
			fibril_mutex_unlock(&term->buffer_mtx);
		}
		window_yield(term->widget.window);
	}
	return damage;
}

static void terminal_rearrange(widget_t *widget, sysarg_t hpos, sysarg_t vpos,
    sysarg_t width, sysarg_t height)
{
	widget_modify(widget, hpos, vpos, width, height);
	widget->width_ideal = width;
	widget->height_ideal = height;

	terminal_t *term = (terminal_t *) widget;

	/* Update mode. */
	fibril_mutex_lock(&term->visualizer->mode_mtx);
	vslmode_t *mode = &term->mode.mode;
	mode->version += 1;
	mode->screen_aspect.width = width;
	mode->screen_aspect.height = height;
	mode->screen_width = width / FONT_WIDTH;
	mode->screen_height = height / FONT_SCANLINES;
	fibril_mutex_unlock(&term->visualizer->mode_mtx);

	/* Allocate new buffer for buffered terminal. */
	if (term->buffered) {
		fibril_mutex_lock(&term->buffer_mtx);
		if (term->buffer) {
			surface_destroy(term->buffer);
		}
		term->buffer = surface_create(width, height, NULL, SURFACE_FLAG_NONE);
		fibril_mutex_unlock(&term->buffer_mtx);
	}

	/* Request console to switch to new visualizer mode. */
	graph_notify_mode_change(term->visualizer->notif_sess,
	    term->visualizer->reg_svc_handle, term->mode.mode.index);

	/* Clear new buffer. */
	if (term->buffered) {
		copy_buffer_internal(term);
	}
}

static void terminal_repaint(widget_t *widget)
{
	terminal_t *term = (terminal_t *) widget;

	if (term->buffered) {
		if (copy_buffer_internal(term)) {
			window_damage(term->widget.window);
		}
	}
}

static void terminal_handle_keyboard_event(widget_t *widget, kbd_event_t event)
{
	terminal_t *term = (terminal_t *) widget;
	async_exch_t *exch = async_exchange_begin(term->relay.sess);
	async_msg_4(exch, INPUT_EVENT_KEY, event.type, event.key, event.mods, event.c);
	async_exchange_end(exch);
}

static void terminal_handle_position_event(widget_t *widget, pos_event_t event)
{
	if (event.type == POS_UPDATE) {
		// TODO - remove this if terminal cursor movement is desirable
		return;
	}

	widget->window->focus = widget;
	terminal_t *term = (terminal_t *) widget;
	async_exch_t *exch = async_exchange_begin(term->relay.sess);

	/* Translate absolute pixel coordinates into relative charfield coordinates.
	 * Because console expects relative position changes, previous position
	 * must be remembered to allow translation from absolute coordinates. */
	ssize_t dcol = ((ssize_t)(event.hpos - widget->hpos) / FONT_WIDTH) -
	    (ssize_t)term->hpos_last;
	ssize_t drow = ((ssize_t)(event.vpos - widget->vpos) / FONT_SCANLINES) -
	    (ssize_t)term->vpos_last;
	term->hpos_last = (event.hpos - widget->hpos) / FONT_WIDTH;
	term->vpos_last = (event.vpos - widget->vpos) / FONT_SCANLINES;

	if (event.type == POS_UPDATE) {
		async_msg_2(exch, INPUT_EVENT_MOVE, dcol, drow);
	} else {
		async_msg_2(exch, INPUT_EVENT_MOVE, dcol, drow);
		async_msg_2(exch, INPUT_EVENT_BUTTON,
		    event.btn_num, (event.type == POS_PRESS));
	}

	async_exchange_end(exch);
}

static void console_connection(ipc_callid_t iid, ipc_call_t *icall, void *arg)
{
	terminal_input_relay_t *relay = input_get_relay(IPC_GET_ARG1(*icall));

	if (relay) {
		ipc_call_t call;
		ipc_callid_t callid;

		/* Accept the connection. */
		async_answer_0(iid, EOK);

		/* Establish callback session. */
		callid = async_get_call(&call);
		relay->sess = async_callback_receive_start(EXCHANGE_SERIALIZE, &call);
		if (relay->sess != NULL) {
			async_answer_0(callid, EOK);
		} else {
			async_answer_0(callid, ELIMIT);
		}
	} else {
		graph_client_connection(iid, icall, arg);
	}
}

static bool spawn_clients(task_id_t console_task, sysarg_t tab_count, bool klog)
{
	int rc;

	/* Wait for console to be ready. */
	int server_idx;
	task_exit_t texit;
	rc = task_wait(console_task, &texit, &server_idx);
	if (rc != EOK || texit != TASK_EXIT_NORMAL) {
		return false;
	}

	/* Spawn console terminals. */
	for (sysarg_t i = 0; i < tab_count; ++i) {
		char virt_cons[LOC_NAME_MAXLEN];
		snprintf(virt_cons, LOC_NAME_MAXLEN, "%s%d%s%d",
		    "term", server_idx, "/vc", (int) i);

		char virt_cons_loc[LOC_NAME_MAXLEN];
		snprintf(virt_cons_loc, LOC_NAME_MAXLEN, "%s/%s",
		    LOCFS_MOUNT_POINT, virt_cons);

		service_id_t service_id;
		rc = loc_service_get_id(virt_cons, &service_id, IPC_FLAG_BLOCKING);
		if (rc != EOK) {
			return false;
		}

		if (klog && (i == (tab_count - 1))) {
			rc = task_spawnl(NULL, APP_GETTERM,
			    APP_GETTERM, virt_cons_loc, APP_KLOG, NULL);
		} else if (i == 0) {
			rc = task_spawnl(NULL, APP_GETTERM,
			    APP_GETTERM, "-w", virt_cons_loc, APP_BDSH, NULL);
		} else {
			rc = task_spawnl(NULL, APP_GETTERM,
			    APP_GETTERM, virt_cons_loc, APP_BDSH, NULL);
		}

		if (rc != EOK) {
			return false;
		}
	}

	return true;
}

bool init_terminal(
    terminal_t *term, widget_t *parent,
    bool buffered, sysarg_t width, sysarg_t height,
    sysarg_t tab_count, bool include_klog)
{
	int rc;

	async_set_client_connection(console_connection);
	rc = loc_server_register(NAME);
	if (rc != EOK) {
		return false;
	}

	rc = graph_find_namespace_idx();
	if (rc != EOK) {
		return false;
	}

	rc = input_find_namespace_idx();
	if (rc != EOK) {
		return false;
	}

	widget_init(&term->widget, parent);

	term->widget.width = width;
	term->widget.height = height;
	term->widget.width_ideal = width;
	term->widget.height_ideal = height;

	term->widget.destroy = terminal_destroy;
	term->widget.reconfigure = terminal_reconfigure;
	term->widget.rearrange = terminal_rearrange;
	term->widget.repaint = terminal_repaint;
	term->widget.handle_keyboard_event = terminal_handle_keyboard_event;
	term->widget.handle_position_event = terminal_handle_position_event;

	term->visualizer = graph_alloc_visualizer();
	if (!term->visualizer) {
		widget_deinit(&term->widget);
		return false;
	}
	graph_init_visualizer(term->visualizer);

	vslmode_t *mode = &term->mode.mode;
	mode->index = 0;
	mode->version = 0;
	mode->refresh_rate = 0;
	mode->screen_aspect.width = width;
	mode->screen_aspect.height = height;
	mode->screen_width = width / FONT_WIDTH;
	mode->screen_height = height / FONT_SCANLINES;
	mode->cell_type = CELL_TYPE_FIELD;
	mode->cell_aspect.width = FONT_WIDTH;
	mode->cell_aspect.height = FONT_SCANLINES;
	mode->cell_visual.field_visual = CHAR_ATTR_RGB;

	link_initialize(&term->mode.link);
	list_append(&term->mode.link, &term->visualizer->modes);

	term->visualizer->def_mode_idx = 0;

	term->visualizer->ops = emu_ops;
	term->visualizer->dev_ctx = term;

	term->buffered = buffered;
	fibril_mutex_initialize(&term->buffer_mtx);
	term->buffer = NULL;
	if (buffered) {
		term->buffer = surface_create(width, height, NULL, SURFACE_FLAG_NONE);
		if (!term->buffer) {
			widget_deinit(&term->widget);
			return false;
		}
	}

	rc = graph_register_visualizer(term->visualizer);
	if (rc != EOK) {
		if (term->buffer) surface_destroy(term->buffer);
		term->visualizer->dev_ctx = NULL;
		list_remove(&term->mode.link);
		graph_destroy_visualizer(term->visualizer);
		widget_deinit(&term->widget);
		return false;
	}

	term->hpos_last = 0;
	term->vpos_last = 0;
	input_init_relay(&term->relay);
	rc = input_register_relay(&term->relay);
	if (rc != EOK) {
		if (term->buffer) surface_destroy(term->buffer);
		graph_unregister_visualizer(term->visualizer);
		term->visualizer->dev_ctx = NULL;
		list_remove(&term->mode.link);
		graph_destroy_visualizer(term->visualizer);
		widget_deinit(&term->widget);
		return false;
	}

	/* Spawn remote console. */
	char *iname = NULL;
	char *oname = NULL;
	loc_service_get_name(term->relay.dsid, &iname);
	loc_service_get_name(term->visualizer->reg_svc_handle, &oname);
	if (iname && oname) {
		rc = task_spawnl(&term->console_task, SRV_CONSOLE, SRV_CONSOLE,
		    iname, oname, NULL);
	} else {
		rc = ENOMEM;
	}
	free(iname);
	free(oname);
	if (rc != EOK) {
		if (term->buffer) surface_destroy(term->buffer);
		input_unregister_relay(&term->relay);
		graph_unregister_visualizer(term->visualizer);
		term->visualizer->dev_ctx = NULL;
		list_remove(&term->mode.link);
		graph_destroy_visualizer(term->visualizer);
		widget_deinit(&term->widget);
		return false;
	}

	/* Spawn console terminals and clients. */
	bool success = spawn_clients(term->console_task, tab_count, include_klog);
	if (!success) {
		task_kill(term->console_task);
		if (term->buffer) surface_destroy(term->buffer);
		input_unregister_relay(&term->relay);
		graph_unregister_visualizer(term->visualizer);
		term->visualizer->dev_ctx = NULL;
		list_remove(&term->mode.link);
		graph_destroy_visualizer(term->visualizer);
		widget_deinit(&term->widget);
		return false;
	}

	return true;
}

terminal_t *create_terminal(widget_t *parent,
    bool buffered, sysarg_t width, sysarg_t height,
    sysarg_t tab_count, bool include_klog)
{
	terminal_t *term = (terminal_t *) malloc(sizeof(terminal_t));
	if (!term) {
		return NULL;
	}

	bool initialized = init_terminal(
	    term, parent, buffered, width, height, tab_count, include_klog);
	if (initialized) {
		return term;
	} else {
		free(term);
		return NULL;
	}
}

/** @}
 */
