/*
 * Copyright (c) 2011 Martin Decky
 * Copyright (c) 2011 Petr Koupy
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup console
 * @{
 */
/** @file
 */

#include <sys/types.h>
#include <bool.h>
#include <errno.h>
#include <str_error.h>
#include <byteorder.h>
#include <stdio.h>

#include <align.h>
#include <as.h>
#include <malloc.h>

#include <atomic.h>
#include <fibril_synch.h>
#include <adt/prodcons.h>
#include <ipc/graph.h>
#include <ipc/input.h>
#include <ipc/console.h>
#include <ipc/vfs.h>

#include <async.h>
#include <loc.h>
#include <devman.h>

#include <event.h>
#include <device/graph_dev.h>
#include <io/keycode.h>
#include <io/mode.h>
#include <io/graph.h>
#include <io/console.h>

#include "scrollgrid.h"
#include "console.h"

#define NAME       "console"
#define NAMESPACE  "term"

#define CONSOLE_COUNT   12
#define KERNEL_CONSOLE  11

#define UTF8_CHAR_BUFFER_SIZE (STR_BOUNDS(1) + 1)

/**
 * Context of a single console tab. Contexts are pre-allocated and registered
 * for all tabs (CONSOLE_COUNT) during server initialization. Their backing
 * storages (scrollgrids) are however allocated lazily upon the connection
 * from client to avoid unnecessary memory overhead.
 */
typedef struct {
	/**
	 * Index of the tab in this instance of console server. */
	size_t index;

	/**
	 * Number of connected clients. Throughout the whole console server,
	 * it is assumed that clients do not compete on a single tab. That is,
	 * they either wait for each other (i.e. parent process waits for
	 * completion of child process), or they cooperate on consuming the
	 * input and damaging the screen. */
	atomic_t ref_cnt;

	/**
	 * System-wide identification of the console tab. */
	service_id_t dsid;

	/**
	 * Mutex guarding the client's internal state that is derived from the
	 * console server state (e.g. scrollgrids dimensions). Usage of this guard
	 * is voluntary for the client, but it is probable that ignoring client
	 * gets corrupted after the terminal was resized. Notifications about
	 * impending and finished changes to the terminal size are currently
	 * masked as special keyboard events (this way it is possible to support
	 * legacy clients). Upon receiving the first special event, client shall
	 * prepare for the change, unlock the mutex and wait for the other event.
	 * Console server locks input during resizing so it is ensured those
	 * special events will come right after each other. When the second
	 * special event is received, the client shall lock the mutex, fetch
	 * new dimensions, restore its state and continue normal operations.*/
	fibril_mutex_t ext_state_mtx;

	/**
	 * Synchronized FIFO queue of keyboard events. Events are produced by
	 * fibril handling the input server notifications and consumed by the
	 * client. Input fibril puts events into the queue of the last active
	 * non-kernel console tab. */
	prodcons_t input_queue;
	
	/** 
	 * Not yet sent bytes of last char event. */
	char char_remains[UTF8_CHAR_BUFFER_SIZE];
	
	/**
	 * Number of not yet sent bytes. */
	size_t char_remains_len;

	/**
	 * Mutex guarding the backing storage access. It is used to mutually
	 * exclude the clients on the single tab (although such scenario
	 * is not practically expected) and to put aside clients when complex
	 * operations are carried out (mode change, tab switch). */
	fibril_mutex_t grid_mtx;

	/**
	 * Lazily allocated backing storage. It is allocated upon first client
	 * connection and then lives until the console server is killed. Backing
	 * storage supports scrolling, cursor/pointer tracking and damage region
	 * tracking. */
	scrollgrid_t *grid;
} console_t;

static console_t consoles[CONSOLE_COUNT];

static console_t *prev_console = &consoles[0];
static console_t *active_console = &consoles[0];
static console_t *kernel_console = &consoles[KERNEL_CONSOLE];

/**
 * Mutex guarding the global console pointers (prev, active), current mode
 * and backup scrollgrid. That way, it is used to mutually exclude infrequent
 * complex operations such as mode change or console tab switch with frequent
 * operations such as input handling. Generally, it synchronizes the
 * notification fibrils from input server and graphic driver. Usage of this
 * mutex by client fibrils is minimized to avoid unnecessary serialization. */
static FIBRIL_MUTEX_INITIALIZE(change_mtx);

/** Session to the input server. Currently used just to establish callback. */
static async_sess_t *input_sess;

/** Session to the visualizer. */
static async_sess_t *vsl_sess;

/** Copy of the current visualizer mode. */
static vslmode_t mode;

/**
 * Each console tab has its private non-shared scrollgrid. Then there is a
 * single global scrollgrid whose charfieldmap is shared with the visualizer
 * to handle damage messages. When a tab becomes active, the content of its
 * private scrollgrid is copied into the shared scrollgrid and the pointers
 * are swapped, that is the active console is using the shared scrollgrid and
 * its original private scrollgrid is stored in this backup_grid pointer until
 * some other tab becomes active. Before the connection of the very first
 * client or when the kconsole is active, the backup_grid pointer is holding
 * the shared scrollgrid. */
static scrollgrid_t *backup_grid;

struct {
	sysarg_t col;
	sysarg_t row;
	
	sysarg_t btn_col;
	sysarg_t btn_row;
	
	bool pressed;
} mouse;

static inline void cons_update(scrollgrid_t *grid)
{
	if (scrollgrid_is_shared(grid)) {
		sysarg_t col, row, cols, rows, top_row;
		scrollgrid_get_damaged_region(grid, &col, &row, &cols, &rows);
		top_row = scrollgrid_get_top_row(grid);
		scrollgrid_reset_damaged_region(grid);
		visualizer_update_damaged_region(
		    vsl_sess, col, row, cols, rows, 0, top_row);
	}
}

static void cons_read(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	ipc_callid_t callid;
	size_t size;
	if (!async_data_read_receive(&callid, &size)) {
		async_answer_0(callid, EINVAL);
		async_answer_0(iid, EINVAL);
		return;
	}

	char *buf = (char *) malloc(size);
	if (buf == NULL) {
		async_answer_0(callid, ENOMEM);
		async_answer_0(iid, ENOMEM);
		return;
	}

	/*
	 * Read input from keyboard and copy it to the buffer.
	 * We need to handle situation when wchar is split by 2 following
	 * reads. 
	 */
	size_t pos = 0;
	while (pos < size) {
		/* Copy to the buffer remaining characters. */
		while ((pos < size) && (cons->char_remains_len > 0)) {
			buf[pos] = cons->char_remains[0];
			pos++;
			/* Unshift the array. */
			for (size_t i = 1; i < cons->char_remains_len; i++) {
				cons->char_remains[i - 1] = cons->char_remains[i];
			}
			cons->char_remains_len--;
		}
		/* Still not enough? Then get another key from the queue. */
		if (pos < size) {
			link_t *link = prodcons_consume(&cons->input_queue);
			kbd_event_t *event = list_get_instance(link, kbd_event_t, link);

			/* Accept key presses of printable chars only. */
			if ((event->type == KEY_PRESS) && (event->c != 0)) {
				wchar_t tmp[2] = { event->c, 0 };
				wstr_to_str(cons->char_remains, UTF8_CHAR_BUFFER_SIZE, tmp);
				cons->char_remains_len = str_size(cons->char_remains);
			}

			free(event);
		}
	}

	async_data_read_finalize(callid, buf, size);
	async_answer_1(iid, EOK, size);
	free(buf);
}

static void cons_write_char(console_t *cons, wchar_t ch)
{
	bool flush = false;

	fibril_mutex_lock(&cons->grid_mtx);

	switch (ch) {
	case '\n':
		flush =  scrollgrid_newline(cons->grid);
		break;
	case '\r':
		break;
	case '\t':
		flush = scrollgrid_tabstop(cons->grid, 8);
		break;
	case '\b':
		flush = scrollgrid_backspace(cons->grid);
		break;
	default:
		flush = scrollgrid_putchar(cons->grid, ch);
	}

	if (flush) {
		cons_update(cons->grid);
	}

	fibril_mutex_unlock(&cons->grid_mtx);
}

static void cons_write(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	void *buf;
	size_t size;
	int rc = async_data_write_accept(&buf, false, 0, 0, 0, &size);

	if (rc != EOK) {
		async_answer_0(iid, rc);
		return;
	}

	size_t off = 0;
	while (off < size) {
		cons_write_char(cons, str_decode(buf, &off, size));
	}

	async_answer_1(iid, EOK, size);
	free(buf);
}

static void cons_sync(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	fibril_mutex_lock(&cons->grid_mtx);
	cons_update(cons->grid);
	fibril_mutex_unlock(&cons->grid_mtx);
	async_answer_0(iid, EOK);
}

static void cons_clear(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	fibril_mutex_lock(&cons->grid_mtx);
	scrollgrid_clear(cons->grid);
	cons_update(cons->grid);
	fibril_mutex_unlock(&cons->grid_mtx);
	async_answer_0(iid, EOK);
}

static void cons_set_cursor(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	fibril_mutex_lock(&cons->grid_mtx);
	scrollgrid_set_cursor(cons->grid, IPC_GET_ARG1(*icall), IPC_GET_ARG2(*icall));
	cons_update(cons->grid);
	fibril_mutex_unlock(&cons->grid_mtx);
	async_answer_0(iid, EOK);
}

static void cons_get_cursor(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	sysarg_t col, row;
	fibril_mutex_lock(&cons->grid_mtx);
	scrollgrid_get_cursor(cons->grid, &col, &row);
	fibril_mutex_unlock(&cons->grid_mtx);
	async_answer_2(iid, EOK, col, row);
}

static void cons_get_size(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	sysarg_t cols, rows;
	fibril_mutex_lock(&cons->grid_mtx);
	scrollgrid_get_dimensions(cons->grid, &cols, &rows);
	fibril_mutex_unlock(&cons->grid_mtx);
	async_answer_2(iid, EOK, cols, rows);
}

static void cons_get_color_cap(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	int rc;
	console_caps_t caps;
	fibril_mutex_lock(&change_mtx);
	switch (mode.cell_visual.field_visual) {
	case CHAR_ATTR_STYLE:
		rc = EOK;
		caps = CONSOLE_CAP_STYLE;
		break;
	case CHAR_ATTR_INDEX:
		rc = EOK;
		caps = CONSOLE_CAP_INDEXED;
		break;
	case CHAR_ATTR_RGB:
		rc = EOK;
		caps = CONSOLE_CAP_RGB;
		break;
	default:
		rc = EINVAL;
		caps = CONSOLE_CAP_NONE;
		break;
	}
	fibril_mutex_unlock(&change_mtx);
	async_answer_1(iid, rc, caps);
}

static void cons_set_style(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	char_attrs_t attrs;
	attrs.type = CHAR_ATTR_STYLE;
	attrs.val.style = IPC_GET_ARG1(*icall);
	fibril_mutex_lock(&cons->grid_mtx);
	scrollgrid_set_default_attrs(cons->grid, attrs);
	fibril_mutex_unlock(&cons->grid_mtx);
	async_answer_0(iid, EOK);
}

static void cons_set_color(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	char_attrs_t attrs;
	attrs.type = CHAR_ATTR_INDEX;
	attrs.val.index.bgcolor = IPC_GET_ARG1(*icall);
	attrs.val.index.fgcolor = IPC_GET_ARG2(*icall);
	attrs.val.index.attr = IPC_GET_ARG3(*icall);
	fibril_mutex_lock(&cons->grid_mtx);
	scrollgrid_set_default_attrs(cons->grid, attrs);
	fibril_mutex_unlock(&cons->grid_mtx);
	async_answer_0(iid, EOK);
}

static void cons_set_rgb_color(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	char_attrs_t attrs;
	attrs.type = CHAR_ATTR_RGB;
	attrs.val.rgb.bgcolor = IPC_GET_ARG1(*icall);
	attrs.val.rgb.fgcolor = IPC_GET_ARG2(*icall);
	fibril_mutex_lock(&cons->grid_mtx);
	scrollgrid_set_default_attrs(cons->grid, attrs);
	fibril_mutex_unlock(&cons->grid_mtx);
	async_answer_0(iid, EOK);
}

static void cons_set_cursor_visibility(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	fibril_mutex_lock(&cons->grid_mtx);
	scrollgrid_set_cursor_visibility(cons->grid, IPC_GET_ARG1(*icall));
	cons_update(cons->grid);
	fibril_mutex_unlock(&cons->grid_mtx);
	async_answer_0(iid, EOK);
}

static void cons_get_event(console_t *cons, ipc_callid_t iid, ipc_call_t *icall)
{
	link_t *link = prodcons_consume(&cons->input_queue);
	kbd_event_t *event = list_get_instance(link, kbd_event_t, link);
	async_answer_4(iid, EOK, event->type, event->key, event->mods, event->c);
	free(event);
}

static void client_connection(ipc_callid_t iid, ipc_call_t *icall, void *arg)
{
	console_t *cons = NULL;

	fibril_mutex_lock(&change_mtx);

	for (size_t i = 0; i < CONSOLE_COUNT; i++) {
		if (i == KERNEL_CONSOLE) {
			continue;
		}
		
		if (consoles[i].dsid == (service_id_t) IPC_GET_ARG1(*icall)) {
			cons = &consoles[i];
			break;
		}
	}
	
	if (cons == NULL) {
		fibril_mutex_unlock(&change_mtx);
		async_answer_0(iid, ENOENT);
		return;
	}

	fibril_mutex_lock(&cons->grid_mtx);

	atomic_count_t client_count = atomic_postinc(&cons->ref_cnt);
	if (client_count == 0) {
		if (cons->grid == NULL) {
			cons->grid = scrollgrid_create(
			    mode.screen_width, mode.screen_height, SCROLLGRID_FLAG_NONE);

			if (cons->grid == NULL) {
				fibril_mutex_unlock(&cons->grid_mtx);
				fibril_mutex_unlock(&change_mtx);
				async_answer_0(iid, ENOMEM);
				return;
			}
			
			if (cons == active_console) {
				scrollgrid_copy_content(cons->grid, backup_grid);
				scrollgrid_t *tmp = cons->grid;
				cons->grid = backup_grid;
				backup_grid = tmp;
			}
		} else {
			scrollgrid_clear(cons->grid);
		}

		scrollgrid_set_cursor_visibility(cons->grid, true);
		scrollgrid_set_pointer_visibility(cons->grid, true);
	}

	cons_update(cons->grid);

	fibril_mutex_unlock(&cons->grid_mtx);
	
	fibril_mutex_unlock(&change_mtx);
	async_answer_0(iid, EOK);
	
	while (true) {
		ipc_call_t call;
		ipc_callid_t callid = async_get_call(&call);
		
		if (!IPC_GET_IMETHOD(call)) {
			async_answer_0(callid, EINVAL);
			break;
		}
		
		switch (IPC_GET_IMETHOD(call)) {
		case VFS_OUT_READ:
			cons_read(cons, callid, &call);
			break;
		case VFS_OUT_WRITE:
			cons_write(cons, callid, &call);
			break;
		case VFS_OUT_SYNC:
			cons_sync(cons, callid, &call);
			break;
		case CONSOLE_CLEAR:
			cons_clear(cons, callid, &call);
			break;
		case CONSOLE_GOTO:
			cons_set_cursor(cons, callid, &call);
			break;
		case CONSOLE_GET_POS:
			cons_get_cursor(cons, callid, &call);
			break;
		case CONSOLE_GET_SIZE:
			cons_get_size(cons, callid, &call);
			break;
		case CONSOLE_GET_COLOR_CAP:
			cons_get_color_cap(cons, callid, &call);
			break;
		case CONSOLE_SET_STYLE:
			cons_set_style(cons, callid, &call);
			break;
		case CONSOLE_SET_COLOR:
			cons_set_color(cons, callid, &call);
			break;
		case CONSOLE_SET_RGB_COLOR:
			cons_set_rgb_color(cons, callid, &call);
			break;
		case CONSOLE_CURSOR_VISIBILITY:
			cons_set_cursor_visibility(cons, callid, &call);
			break;
		case CONSOLE_GET_EVENT:
			cons_get_event(cons, callid, &call);
			break;
		case CONSOLE_LOCK_STATE:
			fibril_mutex_lock(&cons->ext_state_mtx);
			async_answer_0(callid, EOK);
			break;
		case CONSOLE_UNLOCK_STATE:
			fibril_mutex_unlock(&cons->ext_state_mtx);
			async_answer_0(callid, EOK);
			break;
		default:
			async_answer_0(callid, EINVAL);
			goto terminate;
		}
	}

terminate:
	atomic_dec(&cons->ref_cnt);
}

static void cons_mode_change(ipc_callid_t iid, ipc_call_t *icall)
{
	int rc;
	bool allocated;
	sysarg_t mode_idx = IPC_GET_ARG2(*icall);

	/* Among others, stop the input events from being sent to the clients. */
	fibril_mutex_lock(&change_mtx);

	/* Allocate client notification events. */
	kbd_event_t *prolog_events[CONSOLE_COUNT];
	kbd_event_t *epilog_events[CONSOLE_COUNT];
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		prolog_events[i] = NULL;
		epilog_events[i] = NULL;
	}

	allocated = true;
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		if (consoles[i].grid != NULL) {
			prolog_events[i] = (kbd_event_t *) malloc(sizeof(kbd_event_t));
			epilog_events[i] = (kbd_event_t *) malloc(sizeof(kbd_event_t));
			if (prolog_events[i] == NULL || epilog_events[i] == NULL) {
				allocated = false;
			}
		}
	}

	if (!allocated) {
		for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
			free(prolog_events[i]);
			free(epilog_events[i]);
		}
		fibril_mutex_unlock(&change_mtx);
		async_answer_0(iid, ENOMEM);
		return;
	}

	/* Prepare client notification events. */
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		kbd_event_t * event;
		if (consoles[i].grid != NULL) {
			/* Client notifications currently exploit the keyboard event
			 * queue. This way, clients do not need to implement a special
			 * server loop to handle such notifications via async framework.
			 * In order to minimize the risk of conflict with the client that
			 * does not understand such notification, the event is masked as
			 * release of Pause key. */

			event = prolog_events[i];
			link_initialize(&event->link);
			event->type = KEY_RELEASE;
			event->key = KC_PAUSE;
			event->mods = KM_SAVE_STATE;
			event->c = '\0';

			event = epilog_events[i];
			link_initialize(&event->link);
			event->type = KEY_RELEASE;
			event->key = KC_PAUSE;
			event->mods = KM_LOAD_STATE;
			event->c = '\0';
		}
	}

	/* Notify clients about the upcoming change, so they can yield the part
	 * of their internal state that is dependent on the mode (e.g. shell can
	 * temporarily delete the uncommited line from its scrollgrid, so it can
	 * be later restored without any visual artifacts). */
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		if (consoles[i].grid != NULL) {
			prodcons_produce(&consoles[i].input_queue, &prolog_events[i]->link);
		}
	}

	/* Wait until the clients are prepared for the change. */
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		fibril_mutex_lock(&consoles[i].ext_state_mtx);
	}

	/* Stop even clients that did not understand the notification. */
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		fibril_mutex_lock(&consoles[i].grid_mtx);
	}

	/* Retrieve the mode that shall be set. */
	vslmode_t new_mode;
	rc = visualizer_get_mode(vsl_sess, &new_mode, mode_idx);
	if (rc != EOK || new_mode.cell_type != CELL_TYPE_FIELD) {
		for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
			if (consoles[i].grid != NULL) {
				prodcons_produce(&consoles[i].input_queue, &epilog_events[i]->link);
			}
			fibril_mutex_unlock(&consoles[i].ext_state_mtx);
			fibril_mutex_unlock(&consoles[i].grid_mtx);
			
		}
		fibril_mutex_unlock(&change_mtx);
		async_answer_0(iid, EINVAL);
		return;
	}

	/* Before setting the mode, allocate all the required resources for new
	 * scrollgrids. In case of failure, simply rollback and do not change
	 * anything. */
	scrollgrid_t *shared_grid = NULL;
	scrollgrid_t *grids[CONSOLE_COUNT];
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		grids[i] = NULL;
	}

	allocated = true;
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		if (consoles[i].grid != NULL) {
			grids[i] = scrollgrid_create(
				new_mode.screen_width, new_mode.screen_height, SCROLLGRID_FLAG_NONE);
			if (grids[i] == NULL) {
				allocated = false;
			}
		}
	}

	shared_grid = scrollgrid_create(
	    new_mode.screen_width, new_mode.screen_height, SCROLLGRID_FLAG_SHARED);
	if (!shared_grid) {
		allocated = false;
	}

	if (!allocated) {
		for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
			if (grids[i] != NULL) {
				scrollgrid_destroy(grids[i]);
				grids[i] = NULL;
			}
		}
		if (shared_grid != NULL) {
			scrollgrid_destroy(shared_grid);
			shared_grid = NULL;
		}
		for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
			if (consoles[i].grid != NULL) {
				prodcons_produce(&consoles[i].input_queue, &epilog_events[i]->link);
			}
			fibril_mutex_unlock(&consoles[i].ext_state_mtx);
			fibril_mutex_unlock(&consoles[i].grid_mtx);
		}
		fibril_mutex_unlock(&change_mtx);
		async_answer_0(iid, ENOMEM);
		return;
	}

	/* All resources are now prepared. Try to set the new mode. */
	rc = visualizer_set_mode(vsl_sess,
		new_mode.index, new_mode.version, scrollgrid_direct_access(shared_grid));
	if (rc != EOK) {
		for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
			if (grids[i] != NULL) {
				scrollgrid_destroy(grids[i]);
				grids[i] = NULL;
			}
		}
		if (shared_grid != NULL) {
			scrollgrid_destroy(shared_grid);
			shared_grid = NULL;
		}
		for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
			if (consoles[i].grid != NULL) {
				prodcons_produce(&consoles[i].input_queue, &epilog_events[i]->link);
			}
			fibril_mutex_unlock(&consoles[i].ext_state_mtx);
			fibril_mutex_unlock(&consoles[i].grid_mtx);
		}
		fibril_mutex_unlock(&change_mtx);
		async_answer_0(iid, rc);
		return;
	}

	/* Copy old grids to the new grids and deallocate old grids. */
	mode = new_mode;
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		if (&consoles[i] == active_console) {
			if (consoles[i].grid != NULL) {
				/* Active console is visible. Copy both shared and backing
				 * grid and refresh the screen. */

				scrollgrid_copy_content(consoles[i].grid, shared_grid);
				scrollgrid_destroy(consoles[i].grid);
				consoles[i].grid = shared_grid;

				scrollgrid_copy_content(backup_grid, grids[i]);
				scrollgrid_destroy(backup_grid);
				backup_grid = grids[i];

				cons_update(consoles[i].grid);
			} else {
				/* Active console is either kconsole or it is not connected
				 * to any client. Just copy the shared grid. */

				scrollgrid_copy_content(backup_grid, shared_grid);
				scrollgrid_destroy(backup_grid);
				backup_grid = shared_grid;

				cons_update(backup_grid);
			}
		} else {
			if (consoles[i].grid != NULL) {
				scrollgrid_copy_content(consoles[i].grid, grids[i]);
				scrollgrid_destroy(consoles[i].grid);
				consoles[i].grid = grids[i];
			}
		}
	}

	/* Notify clients that they can fetch new information and restore their
	 * state (e.g. shell can restore the uncommited line with the new
	 * dimensions). */
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		if (consoles[i].grid != NULL) {
			prodcons_produce(&consoles[i].input_queue, &epilog_events[i]->link);
		}
		fibril_mutex_unlock(&consoles[i].ext_state_mtx);
		fibril_mutex_unlock(&consoles[i].grid_mtx);
	}
	fibril_mutex_unlock(&change_mtx);

	async_answer_0(iid, EOK);
}

static void cons_visualizer_disconnect(ipc_callid_t iid, ipc_call_t *icall)
{
	fibril_mutex_lock(&change_mtx);
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		fibril_mutex_lock(&consoles[i].grid_mtx);
	}

	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		loc_service_unregister(consoles[i].dsid);
	}

	visualizer_yield(vsl_sess);
	async_hangup(input_sess);
	async_hangup(vsl_sess);

	async_answer_0(iid, EOK);

	/* Do not unlock anything. Whole console server will be soon killed by
	 * the parrent process (e.g. virtual monitor application). */
}

static void vsl_notifications(ipc_callid_t iid, ipc_call_t *icall, void *arg)
{
	/* Ignore parameters, the connection is already opened. */
	while (true) {
		ipc_call_t call;
		ipc_callid_t callid = async_get_call(&call);

		if (!IPC_GET_IMETHOD(call)) {

			async_hangup(vsl_sess);
			return;
		}

		switch (IPC_GET_IMETHOD(call)) {
		case VISUALIZER_MODE_CHANGE:
			cons_mode_change(callid, &call);
			break;
		case VISUALIZER_DISCONNECT:
			cons_visualizer_disconnect(callid, &call);
			return;
		default:
			async_answer_0(callid, EINVAL);
		}
	}
}

static async_sess_t *vsl_connect(const char *svc)
{
	int rc;
	async_sess_t *sess;
	service_id_t dsid;
	devman_handle_t handle;

	rc = loc_service_get_id(svc, &dsid, 0);
	if (rc != EOK) {
		return NULL;
	}

	rc = devman_fun_sid_to_handle(dsid, &handle);
	if (rc == EOK) {
		sess = devman_device_connect(EXCHANGE_SERIALIZE, handle, 0);
		if (sess == NULL) {
			printf("%s: Unable to connect to visualizer %s\n", NAME, svc);
			return NULL;
		}
		rc = graph_dev_connect(sess);
		if (rc != EOK) {
			return NULL;
		}
	} else if (rc == ENOENT) {
		sess = loc_service_connect(EXCHANGE_SERIALIZE, dsid, 0);
		if (sess == NULL) {
			printf("%s: Unable to connect to visualizer %s\n", NAME, svc);
			return NULL;
		}
	} else {
		return NULL;
	}

	async_exch_t *exch = async_exchange_begin(sess);
	rc = async_connect_to_me(exch, 0, 0, 0, vsl_notifications, NULL);
	async_exchange_end(exch);

	if (rc != EOK) {
		async_hangup(sess);
		printf("%s: Unable to create callback connection to service %s (%s)\n",
		    NAME, svc, str_error(rc));
		return NULL;
	}

	return sess;
}

static ssize_t limit(ssize_t val, ssize_t lo, ssize_t hi)
{
	if (val > hi)
		return hi;

	if (val < lo)
		return lo;

	return val;
}

static void cons_mouse_move(sysarg_t dcol, sysarg_t drow)
{
	fibril_mutex_lock(&change_mtx);

	console_t *target_console = active_console;
	if (target_console == kernel_console) {
		target_console = prev_console;
	}

	fibril_mutex_lock(&target_console->grid_mtx);

	if (target_console->grid != NULL) {
		ssize_t scol = (ssize_t) dcol;
		ssize_t srow = (ssize_t) drow;
		sysarg_t cols, rows;
		scrollgrid_get_dimensions(target_console->grid, &cols, &rows);
		mouse.col = limit(mouse.col + scol, 0, cols - 1);
		mouse.row = limit(mouse.row + srow, 0, rows - 1);
		scrollgrid_set_pointer(target_console->grid, mouse.col, mouse.row);
		if (scrollgrid_get_pointer_visibility(target_console->grid) == true) {
			cons_update(target_console->grid);
		}
	}

	fibril_mutex_unlock(&target_console->grid_mtx);

	fibril_mutex_unlock(&change_mtx);
}

static void cons_mouse_button(bool pressed)
{
	fibril_mutex_lock(&change_mtx);

	console_t *target_console = active_console;
	if (target_console == kernel_console) {
		target_console = prev_console;
	}

	fibril_mutex_lock(&target_console->grid_mtx);

	if (target_console->grid != NULL) {
		if (pressed) {
			mouse.btn_col = mouse.col;
			mouse.btn_row = mouse.row;
			mouse.pressed = true;

			// TODO: react to mouse button press (e.g. send event to the client)

			fibril_mutex_unlock(&target_console->grid_mtx);
			fibril_mutex_unlock(&change_mtx);
			return;
		}

		if ((!pressed) && (!mouse.pressed)) {
			fibril_mutex_unlock(&target_console->grid_mtx);
			fibril_mutex_unlock(&change_mtx);
			return;
		}

		mouse.pressed = false;

		// TODO: react to mouse button release (e.g. send event to the client)
	}

	fibril_mutex_unlock(&target_console->grid_mtx);
	fibril_mutex_unlock(&change_mtx);
	return;
}

static void cons_switch(console_t *new_console)
{
	fibril_mutex_lock(&change_mtx);

	bool same_console = (new_console == active_console);
	if (same_console) {
		fibril_mutex_unlock(&change_mtx);
		return;
	}

	fibril_mutex_lock(&active_console->grid_mtx);
	fibril_mutex_lock(&new_console->grid_mtx);

	bool unused_console =
	    new_console->grid == NULL && new_console != kernel_console;
	if (unused_console) {
		fibril_mutex_unlock(&new_console->grid_mtx);
		fibril_mutex_unlock(&active_console->grid_mtx);
		fibril_mutex_unlock(&change_mtx);
		return;
	}

	if (active_console != kernel_console) {
		scrollgrid_copy_content(active_console->grid, backup_grid);
		scrollgrid_t *tmp = active_console->grid;
		active_console->grid = backup_grid;
		backup_grid = tmp;
	}

	if (new_console != kernel_console) {
		scrollgrid_copy_content(new_console->grid, backup_grid);
		scrollgrid_t *tmp = new_console->grid;
		new_console->grid = backup_grid;
		backup_grid = tmp;
		cons_update(new_console->grid);
	} else {
		bool passed = console_kcon();
		if (!passed) {
			scrollgrid_t *tmp = active_console->grid;
			active_console->grid = backup_grid;
			backup_grid = tmp;
		}
	}

	fibril_mutex_unlock(&new_console->grid_mtx);
	fibril_mutex_unlock(&active_console->grid_mtx);

	prev_console = active_console;
	active_console = new_console;

	fibril_mutex_unlock(&change_mtx);
}

static void input_events(ipc_callid_t iid, ipc_call_t *icall, void *arg)
{
	/* Ignore parameters, the connection is already opened. */
	while (true) {
		ipc_call_t call;
		ipc_callid_t callid = async_get_call(&call);

		if (!IPC_GET_IMETHOD(call)) {
			async_hangup(input_sess);
			return;
		}

		kbd_event_type_t type;
		keycode_t key;
		keymod_t mods;
		wchar_t c;

		switch (IPC_GET_IMETHOD(call)) {
		case INPUT_EVENT_KEY:
			type = IPC_GET_ARG1(call);
			key = IPC_GET_ARG2(call);
			mods = IPC_GET_ARG3(call);
			c = IPC_GET_ARG4(call);

			if ((key >= KC_F1) && (key < KC_F1 + CONSOLE_COUNT) &&
			    ((mods & KM_CTRL) == 0)) {
				cons_switch(&consoles[key - KC_F1]);
			} else {
				kbd_event_t *event = (kbd_event_t *) malloc(sizeof(kbd_event_t));
				if (event == NULL) {
					async_answer_0(callid, ENOMEM);
					break;
				}

				link_initialize(&event->link);
				event->type = type;
				event->key = key;
				event->mods = mods;
				event->c = c;

				fibril_mutex_lock(&change_mtx);

				/* Kernel console does not read events from us, so we will
				 * redirect them to the (last) active userspace console
				 * if necessary. */
				console_t *target_console = active_console;
				if (target_console == kernel_console) {
					target_console = prev_console;
				}

				prodcons_produce(&target_console->input_queue, &event->link);

				fibril_mutex_unlock(&change_mtx);
			}

			async_answer_0(callid, EOK);
			break;
		case INPUT_EVENT_MOVE:
			cons_mouse_move(IPC_GET_ARG1(call), IPC_GET_ARG2(call));
			async_answer_0(callid, EOK);
			break;
		case INPUT_EVENT_BUTTON:
			if (IPC_GET_ARG1(call) == 1) {
				cons_mouse_button((bool) IPC_GET_ARG2(call));
			}
			async_answer_0(callid, EOK);
			break;
		default:
			async_answer_0(callid, EINVAL);
		}
	}
}

static async_sess_t *input_connect(const char *svc)
{
	async_sess_t *sess;
	service_id_t dsid;

	int rc = loc_service_get_id(svc, &dsid, 0);
	if (rc == EOK) {
		sess = loc_service_connect(EXCHANGE_ATOMIC, dsid, 0);
		if (sess == NULL) {
			printf("%s: Unable to connect to input service %s\n", NAME, svc);
			return NULL;
		}
	} else
		return NULL;

	async_exch_t *exch = async_exchange_begin(sess);
	rc = async_connect_to_me(exch, 0, 0, 0, input_events, NULL);
	async_exchange_end(exch);

	if (rc != EOK) {
		async_hangup(sess);
		printf("%s: Unable to create callback connection to service %s (%s)\n",
		    NAME, svc, str_error(rc));
		return NULL;
	}

	return sess;
}

static void interrupt_received(ipc_callid_t callid, ipc_call_t *call)
{
	if (active_console == kernel_console) {
		cons_switch(prev_console);
	} else {
		fibril_mutex_lock(&active_console->grid_mtx);
		if (scrollgrid_is_shared(active_console->grid)) {
			sysarg_t cols, rows, top_row;
			scrollgrid_get_dimensions(active_console->grid, &cols, &rows);
			top_row = scrollgrid_get_top_row(active_console->grid);
			visualizer_update_damaged_region(
			    vsl_sess, 0, 0, cols, rows, 0, top_row);
		}
		fibril_mutex_unlock(&active_console->grid_mtx);
	}
}

static int console_srv_init(char *input_svc, char *output_svc)
{
	int rc;

	/* Register console server. */
	async_set_client_connection(client_connection);
	rc = loc_server_register(NAME);
	if (rc != EOK) {
		printf("%s: Unable to register server (%s)\n", NAME, str_error(rc));
		return -1;
	}

	/* Register interrupt handler to switch back from kconsole. */
	async_set_interrupt_received(interrupt_received);
	rc = event_subscribe(EVENT_KCONSOLE, 0);
	if (rc != EOK) {
		printf("%s: Failed to register kconsole notifications (%s)\n",
		    NAME, str_error(rc));
	}

	/* Establish output bidirectional connection. */
	vsl_sess = vsl_connect(output_svc);
	if (vsl_sess == NULL) {
		return -1;
	}

	/* Establish input bidirectional connection. */
	input_sess = input_connect(input_svc);
	if (input_sess == NULL) {
		async_hangup(vsl_sess);
		return -1;
	}

	/* Claim the given visualizer. */
	rc = visualizer_claim(vsl_sess, 0);
	if (rc != EOK) {
		async_hangup(input_sess);
		async_hangup(vsl_sess);
		printf("%s: Unable to claim visualizer (%s)\n", NAME, str_error(rc));
		return -1;
	}

	/* Retrieve the default mode. */
	rc = visualizer_get_default_mode(vsl_sess, &mode);
	if (rc != EOK || mode.cell_type != CELL_TYPE_FIELD) {
		visualizer_yield(vsl_sess);
		async_hangup(input_sess);
		async_hangup(vsl_sess);
		printf("%s: Unable to retrieve mode (%s)\n", NAME, str_error(rc));
		return -1;
	}

	/* Create main backing storage with respect to the retrieved mode. */
	backup_grid = scrollgrid_create(
		mode.screen_width, mode.screen_height, SCROLLGRID_FLAG_SHARED);
	if (backup_grid == NULL) {
		visualizer_yield(vsl_sess);
		async_hangup(input_sess);
		async_hangup(vsl_sess);
		printf("%s: Unable to create grid (%s)\n", NAME, str_error(rc));
		return -1;
	}

	/* Try to set the mode and share out the main backing storage. */
	rc = visualizer_set_mode(vsl_sess,
		mode.index, mode.version, scrollgrid_direct_access(backup_grid));
	if (rc != EOK) {
		visualizer_yield(vsl_sess);
		scrollgrid_destroy(backup_grid);
		async_hangup(input_sess);
		async_hangup(vsl_sess);
		printf("%s: Unable to set mode (%s)\n", NAME, str_error(rc));
		return -1;
	}

	/* Clear the output. */
	cons_update(backup_grid);

	/* Initialize input-specific data structures. */
	scrollgrid_get_pointer(backup_grid, &mouse.col, &mouse.row);
	mouse.pressed = false;

	/* Find unused server index. */
	size_t server_index = 0;
	rc = EEXISTS;
	while (rc == EEXISTS) {
		char probe[LOC_NAME_MAXLEN + 1];
		snprintf(probe, LOC_NAME_MAXLEN, "%s/%zu", NAMESPACE, server_index);
		service_id_t dummy_dsid;

		rc = loc_service_register(probe, &dummy_dsid);
		if (rc != EOK && rc != EEXISTS) {
			visualizer_yield(vsl_sess);
			scrollgrid_destroy(backup_grid);
			async_hangup(input_sess);
			async_hangup(vsl_sess);
			printf("%s: Unable to retrieve server index (%s)\n", NAME, str_error(rc));
			return -1;
		}
		
		if (rc == EEXISTS) {
			/* Server with such index already exists. Increment index. */
			server_index++;
		} else {
			/* Unused index found. */
		}
	}

	/* Initialize and register all consoles. Do not allocate their backing storage yet. */
	for (size_t i = 0; i < CONSOLE_COUNT; ++i) {
		consoles[i].index = i;
		atomic_set(&consoles[i].ref_cnt, 0);
		fibril_mutex_initialize(&consoles[i].ext_state_mtx);
		prodcons_initialize(&consoles[i].input_queue);
		consoles[i].char_remains_len = 0;
		fibril_mutex_initialize(&consoles[i].grid_mtx);
		consoles[i].grid = NULL;
		
		if (i == KERNEL_CONSOLE) {
			continue;
		}
		
		char vc[LOC_NAME_MAXLEN + 1];
		snprintf(vc, LOC_NAME_MAXLEN, "%s%zu/vc%zu", NAMESPACE, server_index, i);
		
		if (loc_service_register(vc, &consoles[i].dsid) != EOK) {
			visualizer_yield(vsl_sess);
			scrollgrid_destroy(backup_grid);
			async_hangup(input_sess);
			async_hangup(vsl_sess);
			for (size_t j = 0; j < i; ++j) {
				loc_service_unregister(consoles[j].dsid);
			}
			printf("%s: Unable to register service %s\n", NAME, vc);
			return -1;
		}
	}

	/* Caller will use server index to discover the services. */
	return server_index;
}

static void usage(void)
{
	printf("Usage: console <input_svc> <output_svc>\n");
}

int main(int argc, char *argv[])
{
	if (argc < 3) {
		usage();
		return -1;
	}
	
	printf("%s: HelenOS Console server\n", NAME);

	int server_id = console_srv_init(argv[1], argv[2]);
	
	if (server_id < 0)
		return -1;
	
	printf("%s: Accepting connections\n", NAME);
	task_retval(server_id);
	async_manager();

	/* Never reached */
	return 0;
}

/** @}
 */

