/*
 * Copyright (c) 2006 Josef Cejka
 * Copyright (c) 2011 Petr Koupy
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup console
 * @{
 */
/** @file
 */

#include <malloc.h>
#include <mem.h>
#include <as.h>
#include <unistd.h>
#include <assert.h>
#include "scrollgrid.h"

/** 
 * Structure for buffering a state of a single tab in the virtual console.
 * Since backing storage is regarded as a circular row buffer, position of
 * a top row is maintained. Fields can be inserted either directly by passing 
 * a fully-specified field or with the help of convenience functions to set a
 * default field attributes and insert characters or other textual primitives.
 * All operations track the bounds of damaged region.
 */
struct scrollgrid {
	scrollgrid_flag_t flags;
	
	sysarg_t cursor_col;
	sysarg_t cursor_row;
	bool cursor_visible;

	sysarg_t pointer_col;
	sysarg_t pointer_row;
	bool pointer_visible;

	sysarg_t dirty_col_lo;
	sysarg_t dirty_col_hi;
	sysarg_t dirty_row_lo;
	sysarg_t dirty_row_hi;

	sysarg_t top_row;

	char_attrs_t cursor_attrs;
	char_attrs_t pointer_attrs;
	char_attrs_t default_attrs;
	
	charfieldmap_t fieldmap;
};

static inline void scrollgrid_insert_into_field_at(
    scrollgrid_t *grid, sysarg_t col, sysarg_t row,
	wchar_t ch, char_attrs_t attrs, bool allow_redirection)
{
	sysarg_t abs_row = (row + grid->top_row) % grid->fieldmap.rows;

	charfield_t *field =
	    charfieldmap_charfield_at(&grid->fieldmap, col, abs_row);
	
	char_attrs_t *field_attrs;
	if (allow_redirection) {
		bool cursor_collision = grid->cursor_visible &&
		    (col == grid->cursor_col) && (row == grid->cursor_row);
		bool pointer_collision = grid->pointer_visible &&
		    (col == grid->pointer_col) && (row == grid->pointer_row);
		bool double_collision = cursor_collision && pointer_collision;

		if (cursor_collision || double_collision) {
			field_attrs = &grid->cursor_attrs;
		} else if (pointer_collision) {
			field_attrs = &grid->pointer_attrs;
		} else {
			field_attrs = &field->attrs;
		}
	} else {
		field_attrs = &field->attrs;
	}

	field->ch = ch;
	*field_attrs = attrs;

	grid->dirty_col_lo = grid->dirty_col_lo > col ? col : grid->dirty_col_lo;
	grid->dirty_col_hi = grid->dirty_col_hi < col ? col : grid->dirty_col_hi;
	grid->dirty_row_lo = grid->dirty_row_lo > row ? row : grid->dirty_row_lo;
	grid->dirty_row_hi = grid->dirty_row_hi < row ? row : grid->dirty_row_hi;
}

static inline void scrollgrid_retrieve_from_field_at(
    scrollgrid_t *grid, sysarg_t col, sysarg_t row,
	wchar_t *ch, char_attrs_t *attrs, bool allow_redirection)
{
	sysarg_t abs_row = (row + grid->top_row) % grid->fieldmap.rows;

	charfield_t *field =
	    charfieldmap_charfield_at(&grid->fieldmap, col, abs_row);
	
	char_attrs_t *field_attrs;
	if (allow_redirection) {
		bool cursor_collision = grid->cursor_visible &&
		    (col == grid->cursor_col) && (row == grid->cursor_row);
		bool pointer_collision = grid->pointer_visible &&
		    (col == grid->pointer_col) && (row == grid->pointer_row);
		bool double_collision = cursor_collision && pointer_collision;

		if (cursor_collision || double_collision) {
			field_attrs = &grid->cursor_attrs;
		} else if (pointer_collision) {
			field_attrs = &grid->pointer_attrs;
		} else {
			field_attrs = &field->attrs;
		}
	} else {
		field_attrs = &field->attrs;
	}

	*ch = field->ch;
	*attrs = *field_attrs;
}

static inline void scrollgrid_clear_row(scrollgrid_t *grid, sysarg_t row)
{
	for (sysarg_t col = 0; col < grid->fieldmap.cols; ++col) {
		scrollgrid_insert_into_field_at(
			grid, col, row, 0, grid->default_attrs, true);
	}
}

static inline void scrollgrid_clear_grid(scrollgrid_t *grid)
{
	for (sysarg_t row = 0; row < grid->fieldmap.rows; ++row) {
		scrollgrid_clear_row(grid, row);
	}
}

scrollgrid_t *scrollgrid_create(sysarg_t cols, sysarg_t rows, scrollgrid_flag_t flags)
{
	scrollgrid_t *grid = (scrollgrid_t *) malloc(sizeof(scrollgrid_t));
	if (grid == NULL) {
		return NULL;
	}

	size_t fieldbuf_size = cols * rows * sizeof(charfield_t);
	charfield_t *fieldbuf;

	if ((flags & SCROLLGRID_FLAG_SHARED) == SCROLLGRID_FLAG_SHARED) {
		fieldbuf = (charfield_t *) as_area_create(AS_AREA_ANY, fieldbuf_size,
		    AS_AREA_READ | AS_AREA_WRITE | AS_AREA_CACHEABLE);
		if (fieldbuf == AS_MAP_FAILED) {
			free(grid);
			return NULL;
		}
	} else {
		fieldbuf = (charfield_t *) malloc(fieldbuf_size);
		if (fieldbuf == NULL) {
			free(grid);
			return NULL;
		}
	}

	grid->flags = flags;

	grid->cursor_col = 0;
	grid->cursor_row = 0;
	grid->cursor_visible = false;

	grid->pointer_col = 0;
	grid->pointer_row = 0;
	grid->pointer_visible = false;

	grid->top_row = 0;

	grid->cursor_attrs.type = CHAR_ATTR_STYLE;
	grid->cursor_attrs.val.style = STYLE_INVERTED;

	grid->pointer_attrs.type = CHAR_ATTR_STYLE;
	grid->pointer_attrs.val.style = STYLE_INVERTED;

	grid->default_attrs.type = CHAR_ATTR_STYLE;
	grid->default_attrs.val.style = STYLE_NORMAL;

	grid->fieldmap.cols = cols;
	grid->fieldmap.rows = rows;
	grid->fieldmap.data = fieldbuf;

	grid->dirty_col_lo = grid->fieldmap.cols - 1;
	grid->dirty_col_hi = 0;
	grid->dirty_row_lo = grid->fieldmap.rows - 1;
	grid->dirty_row_hi = 0;
	scrollgrid_clear_grid(grid);

	return grid;
}

void scrollgrid_destroy(scrollgrid_t *grid)
{
	charfield_t *fieldbuf = grid->fieldmap.data;

	if ((grid->flags & SCROLLGRID_FLAG_SHARED) == SCROLLGRID_FLAG_SHARED)
		as_area_destroy((void *) fieldbuf);
	else
		free(fieldbuf);

	free(grid);
}

static inline void scrollgrid_switch_attributes(scrollgrid_t *grid,
    sysarg_t col, sysarg_t row, char_attrs_t *attrs)
{
	charfield_t field;
	scrollgrid_retrieve_from_field_at(
	    grid, col, row, &field.ch, &field.attrs, false);
	scrollgrid_insert_into_field_at(
	    grid, col, row, field.ch, *attrs, false);
	*attrs = field.attrs;
}

static inline void scrollgrid_update_cursor_visibility(
    scrollgrid_t *grid, bool visible)
{
	if (grid->cursor_visible != visible) {
		bool collision = grid->pointer_visible &&
		    (grid->pointer_col == grid->cursor_col) &&
		    (grid->pointer_row == grid->cursor_row);

		if (collision) {
			/* Leave pointer foreground. */
			char_attrs_t temp;
			temp = grid->cursor_attrs;
			grid->cursor_attrs = grid->pointer_attrs;
			grid->pointer_attrs = temp;
		} else {
			scrollgrid_switch_attributes(
			    grid, grid->cursor_col, grid->cursor_row, &grid->cursor_attrs);
		}

		grid->cursor_visible = visible;
	}
}

static inline void scrollgrid_update_pointer_visibility(
    scrollgrid_t *grid, bool visible)
{
	if (grid->pointer_visible != visible) {
		scrollgrid_switch_attributes(
		    grid, grid->pointer_col, grid->pointer_row, &grid->pointer_attrs);
		grid->pointer_visible = visible;
	}
}

static inline void scrollgrid_update_cursor_pos(
    scrollgrid_t *grid, sysarg_t col, sysarg_t row)
{
	bool cursor_was_visible = grid->cursor_visible;
	scrollgrid_update_cursor_visibility(grid, false);
	grid->cursor_col = col;
	grid->cursor_row = row;
	scrollgrid_update_cursor_visibility(grid, cursor_was_visible);
}

static inline void scrollgrid_update_pointer_pos(
    scrollgrid_t *grid, sysarg_t col, sysarg_t row)
{
	bool pointer_was_visible = grid->pointer_visible;
	scrollgrid_update_pointer_visibility(grid, false);
	grid->pointer_col = col;
	grid->pointer_row = row;
	scrollgrid_update_pointer_visibility(grid, pointer_was_visible);
}

static inline void scrollgrid_inc_cursor_row(scrollgrid_t *grid)
{
	if (grid->cursor_row + 1 == grid->fieldmap.rows) {
		bool cursor_was_visible = grid->cursor_visible;
		bool pointer_was_visible = grid->pointer_visible;
		scrollgrid_update_cursor_visibility(grid, false);
		scrollgrid_update_pointer_visibility(grid, false);

		grid->top_row = (grid->top_row + 1) % grid->fieldmap.rows;
		scrollgrid_clear_row(grid, grid->cursor_row);
		scrollgrid_update_cursor_pos(grid, 0, grid->fieldmap.rows - 1);

		grid->dirty_row_lo = 0;
		grid->dirty_row_hi = grid->fieldmap.rows - 1;

		scrollgrid_update_cursor_visibility(grid, cursor_was_visible);
		scrollgrid_update_pointer_visibility(grid, pointer_was_visible);
	} else {
		scrollgrid_update_cursor_pos(grid, 0, grid->cursor_row + 1);
	}
}

static inline bool scrollgrid_inc_cursor_col(scrollgrid_t *grid)
{
	bool flush = false;
	if (grid->cursor_col + 1 == grid->fieldmap.cols) {
		scrollgrid_inc_cursor_row(grid);
		flush = true;
	} else {
		scrollgrid_update_cursor_pos(
		    grid, grid->cursor_col + 1, grid->cursor_row);
	}

	return flush;
}

static inline void scrollgrid_insert_char_at_cursor(scrollgrid_t *grid, wchar_t ch)
{
	scrollgrid_insert_into_field_at(
	    grid, grid->cursor_col, grid->cursor_row, ch, grid->default_attrs, true);
}

void scrollgrid_copy_content(scrollgrid_t *from, scrollgrid_t *to)
{
	/* Note: Complexity of the following code is caused by the assumption
	 * that the source and destination grids might have different dimensions.
	 * That is, instead of simply copying all fields in each row, the function
	 * gives the destination grid a chance to properly wrap lines and determine
	 * final position of the cursor. */

	bool cursor_was_visible = from->cursor_visible;
	bool pointer_was_visible = from->pointer_visible;
	scrollgrid_update_cursor_visibility(from, false);
	scrollgrid_update_pointer_visibility(from, false);

	to->default_attrs = from->default_attrs;
	to->cursor_attrs = from->cursor_attrs;
	to->pointer_attrs = from->pointer_attrs;
	to->cursor_visible = false;
	to->pointer_visible = false;

	scrollgrid_clear_grid(to);
	scrollgrid_update_cursor_pos(to, 0, 0);
	scrollgrid_update_pointer_pos(to,
	    from->pointer_col < to->fieldmap.cols ? from->pointer_col : to->fieldmap.cols - 1,
	    from->pointer_row < to->fieldmap.rows ? from->pointer_row : to->fieldmap.rows - 1);

	sysarg_t candidate_cursor_col = 0;
	sysarg_t candidate_cursor_row = 0;
	sysarg_t candidate_top_row = 0;
	sysarg_t skipped_rows = 0;
	sysarg_t skipped_cols = 0;
	
	for (sysarg_t row = 0; row < from->fieldmap.rows; ++row) {
		sysarg_t abs_from_row = (row + from->top_row) % from->fieldmap.rows;
		bool something_written = false;

		if (to->cursor_col != 0) {
			scrollgrid_inc_cursor_row(to);
		}

		for (sysarg_t col = 0; col < from->fieldmap.cols; ++col) {
			if (from->cursor_row == row && from->cursor_col == col) {
				while (skipped_rows > 0) {
					scrollgrid_inc_cursor_row(to);
					skipped_rows--;
				}
				while (skipped_cols > 0) {
					scrollgrid_inc_cursor_col(to);
					skipped_cols--;
				}
				candidate_cursor_row = to->cursor_row;
				candidate_cursor_col = to->cursor_col;
				candidate_top_row = to->top_row;
			}

			charfield_t *field =
			    charfieldmap_charfield_at(&from->fieldmap, col, abs_from_row);

			if (field->ch != 0) {
				while (skipped_rows > 0) {
					scrollgrid_inc_cursor_row(to);
					skipped_rows--;
				}
				while (skipped_cols > 0) {
					scrollgrid_inc_cursor_col(to);
					skipped_cols--;
				}
				scrollgrid_insert_into_field_at(to, to->cursor_col, to->cursor_row,
				    field->ch, field->attrs, false);
				scrollgrid_inc_cursor_col(to);
				something_written = true;
			} else {
				skipped_cols++;
			}
		}
		skipped_cols = 0;

		if (!something_written) {
			skipped_rows++;
		}
	}

	/* Put destination grid cursor to such position in the content so it
	 * corresponds to the situation in the source grid. This might not be
	 * possible if the destination grid is smaller (i.e. the desirable
	 * cursor position is scrolled away). Resize-aware applications shall
	 * be prepared for such situation and correct cursor position later
	 * (together with other changes related to resizing). Legacy applications
	 * might get visually corrupted.  */
	sysarg_t diff = to->top_row < candidate_top_row ?
	    candidate_top_row - to->top_row : to->top_row - candidate_top_row;
	candidate_cursor_row =
	    candidate_cursor_row > diff ? candidate_cursor_row - diff : 0;
	scrollgrid_set_cursor(to, candidate_cursor_col, candidate_cursor_row);

	scrollgrid_update_cursor_visibility(to, cursor_was_visible);
	scrollgrid_update_pointer_visibility(to, pointer_was_visible);
}

bool scrollgrid_is_shared(scrollgrid_t * grid)
{
	return ((grid->flags & SCROLLGRID_FLAG_SHARED) == SCROLLGRID_FLAG_SHARED);
}

charfield_t *scrollgrid_direct_access(scrollgrid_t * grid)
{
	return grid->fieldmap.data;
}

void scrollgrid_get_dimensions(scrollgrid_t *grid, sysarg_t *cols, sysarg_t *rows)
{
	assert(cols);
	assert(rows);

	*cols = grid->fieldmap.cols;
	*rows = grid->fieldmap.rows;
}

void scrollgrid_get_damaged_region(scrollgrid_t *grid,
    sysarg_t *col, sysarg_t *row, sysarg_t *cols, sysarg_t *rows)
{
	assert(col);
	assert(row);
	assert(cols);
	assert(rows);

	*col = grid->dirty_col_lo;
	*row = grid->dirty_row_lo;
	*cols = grid->dirty_col_lo <= grid->dirty_col_hi ?
	    (grid->dirty_col_hi - grid->dirty_col_lo) + 1 : 0;
	*rows = grid->dirty_row_lo <= grid->dirty_row_hi ?
	    (grid->dirty_row_hi - grid->dirty_row_lo) + 1 : 0;
}

sysarg_t scrollgrid_get_top_row(scrollgrid_t *grid)
{
	return grid->top_row;
}

void scrollgrid_reset_damaged_region(scrollgrid_t *grid)
{
	grid->dirty_col_lo = grid->fieldmap.cols - 1;
	grid->dirty_col_hi = 0;
	grid->dirty_row_lo = grid->fieldmap.rows - 1;
	grid->dirty_row_hi = 0;
}

void scrollgrid_put_field(scrollgrid_t *grid, sysarg_t col, sysarg_t row, charfield_t field)
{
	scrollgrid_insert_into_field_at(
	    grid, col, row, field.ch, field.attrs, true);
}

charfield_t scrollgrid_get_field(scrollgrid_t *grid, sysarg_t col, sysarg_t row)
{
	charfield_t field;
	scrollgrid_retrieve_from_field_at(
	    grid, col, row, &field.ch, &field.attrs, true);
	return field;
}

void scrollgrid_clear(scrollgrid_t *grid)
{
	scrollgrid_clear_grid(grid);
	scrollgrid_update_cursor_pos(grid, 0, 0);
}

bool scrollgrid_putchar(scrollgrid_t *grid, wchar_t ch)
{
	scrollgrid_insert_char_at_cursor(grid, ch);
	bool flush = scrollgrid_inc_cursor_col(grid);
	return flush;
}

bool scrollgrid_newline(scrollgrid_t *grid)
{
	bool flush = true;
	scrollgrid_inc_cursor_row(grid);
	return flush;
}

bool scrollgrid_tabstop(scrollgrid_t *grid, sysarg_t tab_size)
{
	sysarg_t spaces = tab_size - grid->fieldmap.cols % tab_size;
	bool flush = false;

	for (sysarg_t i = 0; i < spaces; ++i) {
		scrollgrid_insert_char_at_cursor(grid, ' ');
		if (scrollgrid_inc_cursor_col(grid)) {
			flush = true;
		}
	}

	return flush;
}

bool scrollgrid_backspace(scrollgrid_t *grid)
{
	bool flush = false;

	if ((grid->cursor_col == 0) && (grid->cursor_row == 0)) {
		/* Currently no scrollback is supported. */
		// TODO
		flush = false;
	} else if (grid->cursor_col == 0) {
		scrollgrid_update_cursor_pos(
		    grid, grid->fieldmap.cols - 1, grid->cursor_row - 1);
		scrollgrid_insert_char_at_cursor(grid, ' ');
		flush = true;
	} else {
		scrollgrid_update_cursor_pos(
		    grid, grid->cursor_col - 1, grid->cursor_row);
		scrollgrid_insert_char_at_cursor(grid, ' ');
		flush = false;
	}

	return flush;
}

void scrollgrid_set_cursor(scrollgrid_t *grid, sysarg_t col, sysarg_t row)
{
	sysarg_t col_ = col < grid->fieldmap.cols ? col : grid->fieldmap.cols - 1;
	sysarg_t row_ = row < grid->fieldmap.rows ? row : grid->fieldmap.rows - 1;
	scrollgrid_update_cursor_pos(grid, col_, row_);
}

void scrollgrid_get_cursor(scrollgrid_t *grid, sysarg_t *col, sysarg_t *row)
{
	assert(col);
	assert(row);

	*col = grid->cursor_col;
	*row = grid->cursor_row;
}

void scrollgrid_set_cursor_visibility(scrollgrid_t *grid, bool visible)
{
	scrollgrid_update_cursor_visibility(grid, visible);
}

bool scrollgrid_get_cursor_visibility(scrollgrid_t *grid)
{
	return grid->cursor_visible;
}

void scrollgrid_set_pointer(scrollgrid_t *grid, sysarg_t col, sysarg_t row)
{
	sysarg_t col_ = col < grid->fieldmap.cols ? col : grid->fieldmap.cols - 1;
	sysarg_t row_ = row < grid->fieldmap.rows ? row : grid->fieldmap.rows - 1;
	scrollgrid_update_pointer_pos(grid, col_, row_);
}

void scrollgrid_get_pointer(scrollgrid_t *grid, sysarg_t *col, sysarg_t *row)
{
	assert(col);
	assert(row);

	*col = grid->pointer_col;
	*row = grid->pointer_row;
}

void scrollgrid_set_pointer_visibility(scrollgrid_t *grid, bool visible)
{
	scrollgrid_update_pointer_visibility(grid, visible);
}

bool scrollgrid_get_pointer_visibility(scrollgrid_t *grid)
{
	return grid->pointer_visible;
}

void scrollgrid_set_cursor_attrs(scrollgrid_t *grid, char_attrs_t attrs)
{
	bool cursor_was_visible = grid->cursor_visible;
	scrollgrid_update_cursor_visibility(grid, false);

	grid->cursor_attrs = attrs;

	scrollgrid_update_cursor_visibility(grid, cursor_was_visible);
}

void scrollgrid_set_pointer_attrs(scrollgrid_t *grid, char_attrs_t attrs)
{
	bool pointer_was_visible = grid->pointer_visible;
	scrollgrid_update_pointer_visibility(grid, false);

	grid->pointer_attrs = attrs;

	scrollgrid_update_pointer_visibility(grid, pointer_was_visible);
}

void scrollgrid_set_default_attrs(scrollgrid_t *grid, char_attrs_t attrs)
{
	grid->default_attrs = attrs;
}

/** @}
 */
